from typing import List, Optional, Union

import sagepy_connector

from sagepy.core.enzyme import Position, Digest
from sagepy.utility import mass_to_mod

psc = sagepy_connector.py_peptide


class Peptide:
    def __init__(self,
                 decoy: bool,
                 sequence: str,
                 modifications: List[float],
                 mono_isotopic: float,
                 missed_cleavages: int,
                 position: Position,
                 proteins: List[str],
                 semi_enzymatic: bool = False,
                 n_term: Optional[float] = None,
                 c_term: Optional[float] = None
                 ):
        """Peptide class

        Args:
            decoy (bool): Is the peptide a decoy
            sequence (str): The peptide sequence
            modifications (List[float]): The modifications of the peptide
            mono_isotopic (float): The monoisotopic mass of the peptide
            missed_cleavages (int): The number of missed cleavages
            position (Position): The position of the peptide
            proteins (List[str]): The proteins that the peptide is found in
            n_term (Optional[float], optional): Potential modifications on the N-terminal of the peptide. Defaults to None.
            c_term (Optional[float], optional): Potential modifications on the C-terminal of the peptide. Defaults to None.
        """
        self.__peptide_ptr = psc.PyPeptide(decoy, sequence, modifications,
                                           mono_isotopic, missed_cleavages, position.get_py_ptr(),
                                           proteins, semi_enzymatic, n_term, c_term)

    @classmethod
    def from_digest(cls, digest: Digest) -> 'Peptide':
        instance = cls.__new__(cls)
        instance.__peptide_ptr = psc.PyPeptide.try_new_from_digest(digest.get_py_ptr())
        return instance

    @classmethod
    def from_py_peptide(cls, peptide: psc.PyPeptide):
        instance = cls.__new__(cls)
        instance.__peptide_ptr = peptide
        return instance

    @property
    def decoy(self):
        return self.__peptide_ptr.decoy

    @property
    def sequence(self):
        return self.__peptide_ptr.sequence

    @property
    def modifications(self):
        return self.__peptide_ptr.modifications

    @property
    def mono_isotopic(self):
        return self.__peptide_ptr.monoisotopic

    @property
    def missed_cleavages(self):
        return self.__peptide_ptr.missed_cleavages

    @property
    def position(self):
        return Position.from_py_position(self.__peptide_ptr.position)

    @property
    def proteins(self):
        return self.__peptide_ptr.proteins

    @property
    def n_term(self):
        return self.__peptide_ptr.n_term

    @property
    def c_term(self):
        return self.__peptide_ptr.c_term

    @property
    def semi_enzymatic(self):
        return self.__peptide_ptr.semi_enzymatic

    def reverse(self, keep_ends: Union[bool, None]) -> 'Peptide':
        """Reverse the peptide sequence.

        Args:
            keep_ends (Union[bool, None]): Whether to keep the N- and C-terminal amino acids in place.

        Returns:
            Peptide: The reversed peptide.
        """
        return Peptide.from_py_peptide(self.__peptide_ptr.reverse(keep_ends))

    def shuffle(self, keep_ends: Union[bool, None]) -> 'Peptide':
        """Shuffle the peptide sequence.

        Args:
            keep_ends (Union[bool, None]): Whether to keep the N- and C-terminal amino acids in place.

        Returns:
            Peptide: The shuffled peptide.
        """
        return Peptide.from_py_peptide(self.__peptide_ptr.shuffle(keep_ends))

    def get_py_ptr(self):
        return self.__peptide_ptr

    def __repr__(self):
        return f"Peptide(decoy: {self.decoy}, sequence: {self.sequence}, " \
               f"modifications: {self.modifications}, mono_isotopic: {self.mono_isotopic}, " \
               f"missed_cleavages: {self.missed_cleavages}, position: {self.position}, " \
               f"proteins: {self.proteins}, semi_enzymatic: {self.semi_enzymatic}, n_term: {self.n_term}, " \
               f"c_term: {self.c_term})"

    def to_unimod_sequence(self) -> str:
        """ Get Peptide sequence with UNIMOD modification annotations.

        Returns:
            str: Peptide sequence with UNIMOD modification annotations.
        """

        mods = self.modifications
        sequence = self.sequence

        seq = ''

        for i, (s, m) in enumerate(zip(sequence, mods)):
            if m != 0:
                # TODO: check if this is the correct way to handle N- and C-terminal mods
                if i == 0:
                    if mass_to_mod(m) == '[UNIMOD:1]':
                        seq += f'{mass_to_mod(m)}{s}'
                    else:
                        seq += f'{s}{mass_to_mod(m)}'
                else:
                    seq += f'{s}{mass_to_mod(m)}'
            else:
                seq += s

        return seq



from typing import Dict, Tuple, List, Optional

from numba import jit
import numpy as np
import pandas as pd
from numpy.typing import NDArray
from sagepy.qfdr.tdc import target_decoy_competition_pandas

import sagepy_connector

psc = sagepy_connector.py_utility


@jit(nopython=True)
def calculate_ppm_error(measured_value, reference_value):
    ppm_error = ((measured_value - reference_value) / reference_value) * 1_000_000
    return ppm_error


@jit(nopython=True)
def calculate_ppms(measured_values, reference_values):
    n = len(measured_values)
    ppm_errors = np.empty(n, dtype=np.float64)
    for i in range(n):
        ppm_errors[i] = calculate_ppm_error(measured_values[i], reference_values[i])
    return ppm_errors


@jit(nopython=True)
def mean_ppm(mz_observed, mz_calculated) -> float:
    return np.mean(calculate_ppms(mz_observed, mz_calculated))


@jit(nopython=True)
def median_ppm(mz_observed, mz_calculated) -> float:
    return np.median(calculate_ppms(mz_observed, mz_calculated))


def mass_to_mod(mass: float) -> str:
    """ Convert a mass to a UNIMOD modification annotation.

    Args:
        mass: a mass in Da

    Returns:
        a UNIMOD modification annotation
    """
    maybe_key = int(np.round(mass))
    # TODO: find a better way to do the map-back
    mod_dict = {
        42: '[UNIMOD:1]',
        57: '[UNIMOD:4]',
        80: '[UNIMOD:21]',
        16: '[UNIMOD:35]',
        119: '[UNIMOD:312]',
    }
    # try to translate to UNIMOD annotation
    try:
        return mod_dict[maybe_key]
    except KeyError:
        raise KeyError(f"Rounded mass not in dict: {maybe_key}")


def prosit_intensities_to_fragments_map(intensities: NDArray) -> Dict[Tuple[int, int, int], float]:
    """ Convert a Prosit intensity array to a fragment map (ion_type, charge, ordinal) -> intensity.

    Args:
        intensities: a Prosit intensity array

    Returns:
        a fragment map (ion_type, charge, ordinal) -> intensity
    """
    fragment_map = psc.flat_prosit_array_to_fragments_map(intensities)
    return fragment_map


def py_fragments_to_fragments_map(fragments, normalize: bool = True) -> Dict[Tuple[int, int, int], float]:
    """ Convert a Fragments object to a fragment map (ion_type, charge, ordinal) -> intensity.

    Args:
        normalize: whether to normalize the intensities
        fragments: a Fragments object

    Returns:
        a fragment map (ion_type, charge, ordinal) -> intensity
    """
    return psc.py_fragments_to_fragments_map(fragments.get_py_ptr(), normalize)


def peptide_spectrum_match_list_to_pandas(
        psms,
        re_score: bool = False,
        use_sequence_as_match_idx: bool = True) -> pd.DataFrame:
    """Convert a list of peptide spectrum matches to a pandas dataframe

    Args:
        psms (List[PeptideSpectrumMatch]): The peptide spectrum matches
        re_score (bool, optional): Should re-score be used. Defaults to False.
        use_sequence_as_match_idx (bool, optional): Should the sequence be used as the match index. Defaults to True.

    Returns:
        pd.DataFrame: The pandas dataframe
    """
    row_list = []
    for match in psms:
        if match.retention_time_predicted is not None and match.projected_rt is not None:
            delta_rt = match.retention_time_predicted - match.projected_rt
        else:
            delta_rt = None

        if re_score:
            score = match.re_score
        else:
            score = match.hyper_score

        if use_sequence_as_match_idx:
            match_idx = match.sequence
        else:
            match_idx = str(match.peptide_idx)

        if match.inverse_mobility_predicted is not None:
            delta_ims = match.inverse_mobility_predicted - match.inverse_mobility_observed
        else:
            delta_ims = None

        if match.beta_score is not None:
            beta_score = match.beta_score
        else:
            beta_score = None

        row_list.append({
            "spec_idx": match.spec_idx,
            "match_idx": match_idx,
            "proteins": match.proteins,
            "decoy": match.decoy,
            "score": score,
            "re_score": match.re_score,
            "hyper_score": match.hyper_score,
            "rank": match.rank,
            "mono_mz_calculated": match.mono_mz_calculated,
            "mono_mass_observed": match.mono_mass_observed,
            "mono_mass_calculated": match.mono_mass_calculated,
            "delta_mass": match.mono_mass_calculated - match.mono_mass_observed,
            "isotope_error": match.isotope_error,
            "average_ppm": match.average_ppm,
            "delta_next": match.delta_next,
            "delta_best": match.delta_best,
            "matched_peaks": match.matched_peaks,
            "longest_b": match.longest_b,
            "longest_y": match.longest_y,
            "longest_y_pct": match.longest_y_pct,
            "missed_cleavages": match.missed_cleavages,
            "matched_intensity_pct": match.matched_intensity_pct,
            "scored_candidates": match.scored_candidates,
            "poisson": match.poisson,
            "sequence": match.sequence,
            "charge": match.charge,
            "retention_time_observed": match.retention_time_observed,
            "retention_time_predicted": match.retention_time_predicted,
            "delta_rt": delta_rt,
            "inverse_mobility_observed": match.inverse_mobility_observed,
            "inverse_mobility_predicted": match.inverse_mobility_predicted,
            "delta_ims": delta_ims,
            "intensity_ms1": match.intensity_ms1,
            "intensity_ms2": match.intensity_ms2,
            "q_value": match.q_value,
            "collision_energy": match.collision_energy,
            "cosine_similarity": match.cosine_similarity,
            "mean_ppm": mean_ppm(match.fragments_observed.mz_experimental, match.fragments_observed.mz_calculated),
            "median_ppm": median_ppm(match.fragments_observed.mz_experimental, match.fragments_observed.mz_calculated),
            "fragments_observed": match.fragments_observed,
            "fragments_predicted": match.fragments_predicted,
            "projected_rt": match.projected_rt,
            "beta_score": beta_score
        })

    return pd.DataFrame(row_list)


def get_features(ds: pd.DataFrame, score: Optional[str] = None) -> (NDArray, NDArray):

    score = score if score is not None else "score"

    features = [
        f"{score}", "delta_rt", "delta_ims", "cosine_similarity", "delta_mass",
        "rank", "isotope_error", "average_ppm", "delta_next", "delta_best",
        "matched_peaks", "longest_b", "longest_y", "longest_y_pct", "missed_cleavages",
        "matched_intensity_pct", "poisson", "charge",
        "intensity_ms1", "intensity_ms2", "collision_energy"
    ]
    ds = ds.copy()

    ds["intensity_ms1"] = ds["intensity_ms1"].apply(lambda x: np.log(x + 1))
    ds["intensity_ms2"] = ds["intensity_ms2"].apply(lambda x: np.log(x + 1))

    X = ds[features].to_numpy().astype(np.float32)
    Y = ds["decoy"].to_numpy()
    Y = np.array([0 if x else 1 for x in Y]).astype(np.float32)

    return X, Y


def apply_mz_calibration(psm, fragments: pd.DataFrame, use_median: bool = True,
                         target_q: float = 0.01) -> float:
    """Apply a mass calibration to a set of peptide spectrum matches

    Args:
        psm: The peptide spectrum matches
        fragments: The fragments
        use_median: Whether to use the median ppm error
        target_q: The target q-value

    Returns:
        float: The ppm error
    """

    psms = []
    for _, item in psm.items():
        psms.extend(item)

    P = peptide_spectrum_match_list_to_pandas(psms)
    TDC = target_decoy_competition_pandas(P)
    TDC = TDC[TDC.q_value <= target_q]

    B = pd.merge(P, TDC, on=["spec_idx", "match_idx"])

    ppm_error = 0.0

    if use_median:
        ppm_error = np.median(B.median_ppm)
    else:
        ppm_error = np.mean(B.mean_ppm)

    fragments.apply(lambda row: row.processed_spec.calibrate_mz_ppm(ppm_error), axis=1)

    return ppm_error



from typing import Optional, List

import sagepy_connector

from sagepy.core import ProcessedSpectrum
from sagepy.core.scoring import Feature
from sagepy.core.spectrum import Peak

psc = sagepy_connector.py_tmt


class Isobaric:
    def __init__(self, type_name: str):
        types = ["tmt6", "tmt10", "tmt11", "tmt16", "tmt18"]
        if type_name in types:
            self.__isobaric_ptr = psc.PyIsobaric(type_name)
        else:
            raise ValueError(f"Invalid isobaric type, allowed values are: {types}")

    @classmethod
    def from_py_isobaric(cls, isobaric: psc.PyIsobaric):
        instance = cls.__new__(cls)
        instance.__isobaric_ptr = isobaric
        return instance

    @property
    def type_name(self):
        return self.__isobaric_ptr.type_name

    def __repr__(self):
        return f"Isobaric({self.__isobaric_ptr.type_name})"

    def get_py_ptr(self):
        return self.__isobaric_ptr

    def modification_mass(self) -> Optional[float]:
        maybe_mass = self.__isobaric_ptr.modification_mass()
        if maybe_mass is None:
            return None
        return maybe_mass


class Purity:
    def __init__(self, ratio: float, correct_precursors: int, incorrect_precursors: int):
        self.__purity_ptr = psc.PyPurity(ratio, correct_precursors, incorrect_precursors)

    @classmethod
    def from_py_purity(cls, purity: psc.PyPurity):
        instance = cls.__new__(cls)
        instance.__purity_ptr = purity
        return instance

    @property
    def ratio(self):
        return self.__purity_ptr.ratio

    @property
    def correct_precursors(self):
        return self.__purity_ptr.correct_precursors

    @property
    def incorrect_precursors(self):
        return self.__purity_ptr.incorrect_precursors

    def __repr__(self):
        return (f"Purity(ratio={self.ratio}, correct_precursors={self.correct_precursors}, "
                f"incorrect_precursors={self.incorrect_precursors})")

    def get_py_ptr(self):
        return self.__purity_ptr


class Quant:
    def __init__(self, hit: Feature, hit_purity: Purity, spectrum: ProcessedSpectrum,
                 chimera: Optional[Feature] = None, chimera_purity: Optional[Purity] = None,
                 intensities: Optional[List[Peak]] = None):

        if chimera is not None:
            chimera = chimera.get_py_ptr()

        if chimera_purity is not None:
            chimera_purity = chimera_purity.get_py_ptr()

        if intensities is not None:
            intensities = [peak.get_py_ptr() for peak in intensities]

        self.__quant_ptr = psc.PyQuant(hit.get_py_ptr(), hit_purity.get_py_ptr(),
                                       spectrum.get_py_ptr(), chimera, chimera_purity, intensities)

    @classmethod
    def from_py_quant(cls, quant: psc.PyQuant):
        instance = cls.__new__(cls)
        instance.__quant_ptr = quant
        return instance

    @property
    def hit(self):
        return Feature.from_py_feature(self.__quant_ptr.hit())

    @property
    def hit_purity(self):
        return Purity.from_py_purity(self.__quant_ptr.hit_purity())

    @property
    def spectrum(self):
        return ProcessedSpectrum.from_py_processed_spectrum(self.__quant_ptr.spectrum())

    @property
    def chimera(self):
        maybe_chimera = self.__quant_ptr.chimera()
        if maybe_chimera is None:
            return None
        return Feature.from_py_feature(maybe_chimera)

    @property
    def chimera_purity(self):
        maybe_chimera_purity = self.__quant_ptr.chimera_purity()
        if maybe_chimera_purity is None:
            return None
        return Purity.from_py_purity(maybe_chimera_purity)

    @property
    def intensities(self):
        intensities = self.__quant_ptr.intensities()
        if intensities is None:
            return None
        return [Peak.from_py_peak(peak) for peak in intensities]

    def __repr__(self):
        return (f"Quant(hit={self.hit}, hit_purity={self.hit_purity}, spectrum={self.spectrum}, "
                f"chimera={self.chimera}, chimera_purity={self.chimera_purity}, intensities={self.intensities})")

    def get_py_ptr(self):
        return self.__quant_ptr



import numpy as np

from typing import List, Optional

import sagepy_connector
from numpy.typing import NDArray

from sagepy.core.mass import Tolerance
psc = sagepy_connector.py_spectrum


class Peak:
    def __init__(self, mass: float, intensity: float):
        """Peak class

        Args:
            mass (float): The mass of the peak
            intensity (float): The intensity of the peak
        """
        self.__peak_ptr = psc.PyPeak(mass, intensity)

    @classmethod
    def from_py_peak(cls, peak: psc.PyPeak):
        instance = cls.__new__(cls)
        instance.__peak_ptr = peak
        return instance

    @property
    def mass(self):
        return self.__peak_ptr.mass

    @property
    def intensity(self):
        return self.__peak_ptr.intensity

    def get_py_ptr(self):
        return self.__peak_ptr

    def __repr__(self):
        return f"Peak(mass: {self.mass}, intensity: {self.intensity})"


class Deisotoped:
    def __init__(self, mz: float, intensity: float, charge: int = None, envelope: int = None):
        """Deisotoped class

        Args:
            mz (float): The mz of the peak
            intensity (float): The intensity of the peak
            charge (int, optional): The charge of the peak. Defaults to None.
            envelope (int, optional): The envelope of the peak. Defaults to None.
        """
        self.__deisotoped_ptr = psc.PyDeisotoped(mz, intensity, charge, envelope)

    @classmethod
    def from_py_deisotoped(cls, deisotoped: psc.PyDeisotoped):
        instance = cls.__new__(cls)
        instance.__deisotoped_ptr = deisotoped
        return instance

    @property
    def mz(self):
        return self.__deisotoped_ptr.mz

    @property
    def intensity(self):
        return self.__deisotoped_ptr.intensity

    @property
    def charge(self):
        return self.__deisotoped_ptr.charge

    @property
    def envelope(self):
        return self.__deisotoped_ptr.envelope

    def get_py_ptr(self):
        return self.__deisotoped_ptr

    def __repr__(self):
        return f"Deisotoped(mz: {self.mz}, intensity: {self.intensity}, charge: {self.charge}, envelope: {self.envelope})"


class Precursor:
    def __init__(self, mz: float, intensity: float = None, charge: int = None,
                 spectrum_ref: str = None, isolation_window: Tolerance = None,
                 inverse_ion_mobility: float = None, collision_energy: Optional[float] = None):
        """Precursor class

        Args:
            mz (float): The mz of the precursor
            intensity (float, optional): The intensity of the precursor. Defaults to None.
            charge (int, optional): The charge of the precursor. Defaults to None.
            spectrum_ref (str, optional): The spectrum reference of the precursor. Defaults to None.
            isolation_window (Tolerance, optional): The isolation window of the precursor. Defaults to None.
            inverse_ion_mobility (float, optional): The inverse ion mobility of the precursor. Defaults to None.
        """
        if isolation_window is not None:
            self.__precursor_ptr = psc.PyPrecursor(mz, intensity, charge,
                                                   spectrum_ref, isolation_window.get_py_ptr(), inverse_ion_mobility,
                                                   collision_energy)
        else:
            self.__precursor_ptr = psc.PyPrecursor(mz, intensity, charge, spectrum_ref, None, inverse_ion_mobility,
                                                   collision_energy)

    @classmethod
    def from_py_precursor(cls, precursor: psc.PyPrecursor):
        instance = cls.__new__(cls)
        instance.__precursor_ptr = precursor
        return instance

    @property
    def mz(self):
        return self.__precursor_ptr.mz

    def calibrate_mz_ppm(self, ppm: float):
        self.__precursor_ptr.calibrate_mz_ppm(ppm)

    @property
    def intensity(self):
        return self.__precursor_ptr.intensity

    @property
    def charge(self):
        return self.__precursor_ptr.charge

    @property
    def spectrum_ref(self):
        return self.__precursor_ptr.spectrum_ref

    @property
    def isolation_window(self):
        is_window_ptr = self.__precursor_ptr.isolation_window
        if is_window_ptr is not None:
            return Tolerance.from_py_tolerance(is_window_ptr)
        else:
            return None

    @property
    def inverse_ion_mobility(self):
        return self.__precursor_ptr.inverse_ion_mobility

    @property
    def collision_energy(self):
        return self.__precursor_ptr.collision_energy

    def get_py_ptr(self):
        return self.__precursor_ptr

    def __repr__(self):
        return (f"Precursor(mz: {np.round(self.mz, 2)}, "
                f"intensity: {self.intensity}, "
                f"charge: {self.charge}, "
                f"spectrum_ref: {self.spectrum_ref}, "
                f"isolation_window: {self.isolation_window}), "
                f"inverse_ion_mobility: {np.round(self.inverse_ion_mobility, 2) if self.inverse_ion_mobility is not None else self.inverse_ion_mobility}, "
                f"collision_energy: {np.round(self.collision_energy, 2) if self.collision_energy is not None else self.collision_energy})")


class Representation:
    def __init__(self, representation: str = 'centroid'):
        """Representation class

        Args:
            representation (str, optional): The representation of the spectrum. Defaults to 'centroid'.
        """
        self.__representation_ptr = psc.PyRepresentation(representation)

    @classmethod
    def from_py_representation(cls, representation: psc.PyRepresentation):
        instance = cls.__new__(cls)
        instance.__representation_ptr = representation
        return instance

    @property
    def representation(self):
        return self.__representation_ptr.representation_as_string

    def get_py_ptr(self):
        return self.__representation_ptr

    def __repr__(self):
        return f"Representation(representation: {self.representation})"


class ProcessedSpectrum:
    def __init__(self,
                 id: str,
                 level: int,
                 file_id: int,
                 scan_start_time: float,
                 ion_injection_time: float,
                 precursors: List[Precursor],
                 peaks: List[Peak],
                 total_ion_current: float):
        """ProcessedSpectrum class

        Args:
            id (str): The id of the spectrum
            level (int): The level of the spectrum
            file_id (int): The file id of the spectrum
            scan_start_time (float): The scan start time of the spectrum
            ion_injection_time (float): The ion injection time of the spectrum
            precursors (List[Precursor]): The precursors of the spectrum
            peaks (List[Peak]): The peaks of the spectrum
            total_ion_current (float): The total ion current of the spectrum
        """
        self.__processed_spectrum_ptr = psc.PyProcessedSpectrum(
            level, id, file_id, scan_start_time,
            ion_injection_time, [p.get_py_ptr() for p in precursors],
            [p.get_py_ptr() for p in peaks], total_ion_current)

    @classmethod
    def from_py_processed_spectrum(cls, processed_spectrum: psc.PyProcessedSpectrum):
        instance = cls.__new__(cls)
        instance.__processed_spectrum_ptr = processed_spectrum
        return instance

    @property
    def level(self):
        return self.__processed_spectrum_ptr.level

    @property
    def id(self):
        return self.__processed_spectrum_ptr.id

    @property
    def file_id(self):
        return self.__processed_spectrum_ptr.file_id

    @property
    def scan_start_time(self):
        return self.__processed_spectrum_ptr.scan_start_time

    @property
    def ion_injection_time(self):
        return self.__processed_spectrum_ptr.ion_injection_time

    @property
    def precursors(self):
        return [Precursor.from_py_precursor(p) for p in self.__processed_spectrum_ptr.precursors]

    @property
    def peaks(self):
        return [Peak.from_py_peak(p) for p in self.__processed_spectrum_ptr.peaks]

    @property
    def total_ion_current(self):
        return self.__processed_spectrum_ptr.total_ion_current

    @property
    def collision_energies(self):
        return self.__processed_spectrum_ptr.collision_energies

    def calibrate_mz_ppm(self, ppm: float, calibrate_precursors: bool = True):
        if calibrate_precursors:
            for precursor in self.precursors:
                precursor.calibrate_mz_ppm(ppm)
        self.__processed_spectrum_ptr.calibrate_mz_ppm(ppm)

    def get_py_ptr(self):
        return self.__processed_spectrum_ptr

    def __repr__(self):
        return (f"ProcessedSpectrum(level: {self.level}, "
                f"id: {self.id}, "
                f"file_id: {self.file_id}, "
                f"scan_start_time: {np.round(self.scan_start_time, 2)}, "
                f"ion_injection_time: {np.round(self.ion_injection_time, 2)}, "
                f"num_precursors: {len(self.precursors)}, "
                f"num_peaks: {len(self.peaks)}, "
                f"collision_energies: {self.collision_energies},"
                f"total_ion_current: {self.total_ion_current})")


class RawSpectrum:
    def __init__(self,
                 file_id: int,
                 spec_id: str,
                 total_ion_current: float,
                 precursors: List[Precursor],
                 mz: NDArray[float],
                 intensity: NDArray[float],
                 representation: Representation = Representation(),
                 scan_start_time: float = 0.0,
                 ion_injection_time: float = 0.0,
                 ms_level: int = 2,
                 ):
        """RawSpectrum class

        Args:
            file_id (int): The file id of the spectrum
            spec_id (str): The id of the spectrum
            total_ion_current (float): The total ion current of the spectrum
            precursors (List[Precursor]): The precursors of the spectrum
            mz (NDArray[float]): The mz of the peaks of the spectrum
            intensity (NDArray[float]): The intensity of the peaks of the spectrum
            representation (Representation, optional): The representation of the spectrum. Defaults to Representation().
            scan_start_time (float, optional): The scan start time of the spectrum. Defaults to 0.0.
            ion_injection_time (float, optional): The ion injection time of the spectrum. Defaults to 0.0.
            ms_level (int, optional): The ms level of the spectrum. Defaults to 2.
        """
        self.__raw_spectrum_ptr = psc.PyRawSpectrum(file_id, ms_level, spec_id, [p.get_py_ptr() for p in precursors],
                                                    representation.get_py_ptr(),
                                                    scan_start_time, ion_injection_time, total_ion_current,
                                                    mz, intensity)
    @classmethod
    def from_py_raw_spectrum(cls, raw_spectrum: psc.PyRawSpectrum):
        instance = cls.__new__(cls)
        instance.__raw_spectrum_ptr = raw_spectrum
        return instance

    @property
    def ms_level(self):
        return self.__raw_spectrum_ptr.ms_level

    @property
    def id(self):
        return self.__raw_spectrum_ptr.id

    @property
    def precursors(self):
        return [Precursor.from_py_precursor(p) for p in self.__raw_spectrum_ptr.precursors]

    @property
    def representation(self):
        return Representation.from_py_representation(self.__raw_spectrum_ptr.representation)

    @property
    def scan_start_time(self):
        return self.__raw_spectrum_ptr.scan_start_time

    @property
    def ion_injection_time(self):
        return self.__raw_spectrum_ptr.ion_injection_time

    @property
    def total_ion_current(self):
        return self.__raw_spectrum_ptr.total_ion_current

    @property
    def mz(self) -> NDArray[float]:
        return self.__raw_spectrum_ptr.mz

    @property
    def intensity(self) -> NDArray[float]:
        return self.__raw_spectrum_ptr.intensity

    def get_py_ptr(self):
        return self.__raw_spectrum_ptr

    def filter_top_n_peaks(self, n: int) -> 'RawSpectrum':
        return RawSpectrum.from_py_raw_spectrum(self.__raw_spectrum_ptr.filter_top_n_peaks(n))

    def __repr__(self):
        return (f"RawSpectrum(ms_level: {self.ms_level}, "
                f"id: {self.id}, "
                f"precursors: {self.precursors}, "
                f"representation: {self.representation}, "
                f"scan_start_time: {self.scan_start_time}, "
                f"ion_injection_time: {self.ion_injection_time}, "
                f"total_ion_current: {self.total_ion_current}, "
                f"num_peaks: {len(self.mz)})")


class SpectrumProcessor:
    def __init__(
            self, take_top_n: int = 150,
            min_fragment_mz: float = 150,
            max_fragment_mz: float = 2000,
            min_deisotope_mz: float = 0.0,
            deisotope: bool = False,
    ):
        """SpectrumProcessor class

        Args:
            take_top_n (int, optional): The number of peaks to take. Defaults to 150.
            min_fragment_mz (float, optional): The minimum fragment mz. Defaults to 150.
            max_fragment_mz (float, optional): The maximum fragment mz. Defaults to 2000.
            min_deisotope_mz (float, optional): The minimum deisotope mz. Defaults to 0.0.
            deisotope (bool, optional): Whether to deisotope the spectrum. Defaults to False.
        """
        self.__spectrum_processor_ptr = psc.PySpectrumProcessor(
            take_top_n, max_fragment_mz, min_fragment_mz, min_deisotope_mz, deisotope)

    @classmethod
    def from_py_spectrum_processor(cls, spectrum_processor: psc.PySpectrumProcessor):
        instance = cls.__new__(cls)
        instance.__spectrum_processor_ptr = spectrum_processor
        return instance

    def get_py_ptr(self):
        return self.__spectrum_processor_ptr

    @property
    def take_top_n(self):
        return self.__spectrum_processor_ptr.take_top_n

    @property
    def max_fragment_mz(self):
        return self.__spectrum_processor_ptr.max_fragment_mz

    @property
    def min_fragment_mz(self):
        return self.__spectrum_processor_ptr.min_fragment_mz

    @property
    def deisotope(self):
        return self.__spectrum_processor_ptr.deisotope

    def __repr__(self):
        return f"SpectrumProcessor(take_top_n: {self.take_top_n}, max_fragment_mz: {self.max_fragment_mz}, " \
               f"min_fragment_mz: {self.min_fragment_mz}, deisotope: {self.deisotope})"

    def process(self, raw_spectrum: RawSpectrum) -> ProcessedSpectrum:
        return ProcessedSpectrum.from_py_processed_spectrum(self.__spectrum_processor_ptr.process(raw_spectrum.get_py_ptr()))



from .enzyme import EnzymeParameters, Digest
import sagepy_connector
psc = sagepy_connector.py_fasta


class Fasta:
    def __init__(self, fasta_str: str, decoy_tag: str = 'decoy_', generate_decoys: bool = False):
        """Fasta class

        Args:
            fasta_str (str): The fasta string
            decoy_tag (str, optional): The decoy tag. Defaults to 'decoy_'.
            generate_decoys (bool, optional): Should decoys be generated. Defaults to False.
        """
        self.__fasta_ptr = psc.PyFasta.parse(fasta_str, decoy_tag, generate_decoys)

    @classmethod
    def from_py_fasta(cls, fasta: psc.PyFasta):
        instance = cls.__new__(cls)
        instance.__fasta_ptr = fasta
        return instance

    def _digest(self, enzyme_parameters: EnzymeParameters):
        return [Digest.from_py_digest(s) for s in self.__fasta_ptr.digest(enzyme_parameters.get_py_ptr())]



from typing import Dict, List

import sagepy_connector

psc = sagepy_connector.py_modification


class SAGE_KNOWN_MODS:
    @staticmethod
    def n_terminal_static():
        return "^", 304.207

    @staticmethod
    def lysine_static():
        return "K", 304.207

    @staticmethod
    def cysteine_static():
        return "C", 57.0215

    @staticmethod
    def phospho_serine_static():
        return "S", 79.9663

    @staticmethod
    def phospho_threonine_static():
        return "T", 79.9663

    @staticmethod
    def phospho_tyrosine_static():
        return "Y", 79.9663

    @staticmethod
    def phospho_serine_variable():
        return "S", [79.9663]

    @staticmethod
    def phospho_threonine_variable():
        return "T", [79.9663]

    @staticmethod
    def phospho_tyrosine_variable():
        return "Y", [79.9663]

    @staticmethod
    def methionine_variable():
        return "M", [15.9949]

    @staticmethod
    def q_variable():
        return "^Q", [-17.026549]

    @staticmethod
    def glutamic_acid_n_terminal_variable():
        return "^E", [-18.010565]

    @staticmethod
    def peptide_c_terminal_variable():
        return "$", [49.2, 22.9]

    @staticmethod
    def protein_n_terminus_variable():
        return "[", [42.0]

    @staticmethod
    def protein_c_terminal_variable():
        return "]", [111.0]

    def __repr__(self):
        return (f"SAGE_KNOWN_MODS({self.n_terminal_static()}, "
                f"{self.lysine_static()}, "
                f"{self.cysteine_static()}, "
                f"{self.methionine_variable()}, "
                f"{self.q_variable()}, "
                f"{self.glutamic_acid_n_terminal_variable()}, "
                f"{self.peptide_c_terminal_variable()}, "
                f"{self.protein_n_terminus_variable()}, "
                f"{self.protein_c_terminal_variable()})")


# TODO: need to re-implement based on constant modification list
class ModificationSpecificity:
    def __init__(self, s: str):
        self.__modification_specificity_ptr = psc.PyModificationSpecificity(s)

    @classmethod
    def from_py_modification_specificity(cls, specificity: psc.PyModificationSpecificity):
        instance = cls.__new__(cls)
        instance.__modification_specificity_ptr = specificity
        return instance

    def __repr__(self):
        return f"ModificationSpecificity({self.__modification_specificity_ptr.as_string})"

    def get_py_ptr(self):
        return self.__modification_specificity_ptr


def validate_mods(mods: Dict[str, float]) -> Dict[ModificationSpecificity, float]:

    py_validate_dict = psc.py_validate_mods(mods)

    return {ModificationSpecificity.from_py_modification_specificity(k): v for k, v in py_validate_dict.items()}


def validate_var_mods(mods: Dict[str, List[float]]) -> Dict[ModificationSpecificity, List[float]]:

    py_validate_dict = psc.py_validate_var_mods(mods)

    return {ModificationSpecificity.from_py_modification_specificity(k): v for k, v in py_validate_dict.items()}


if __name__ == "__main__":
    static_mods = {k: v for k, v in [SAGE_KNOWN_MODS.cysteine_static()]}
    variable_mods = {k: v for k, v in [SAGE_KNOWN_MODS.methionine_variable()]}

    static = validate_mods(static_mods)
    variab = validate_var_mods(variable_mods)



from typing import Optional, List, Union, Tuple, Dict

import pandas as pd
import sagepy_connector
from .spectrum import ProcessedSpectrum

psc = sagepy_connector.py_scoring
psc_utils = sagepy_connector.py_utility
from .ion_series import IonType
from .mass import Tolerance
from .database import PeptideIx, IndexedDatabase


class ScoreType:
    def __init__(self, name: str):
        name = name.lower()
        names = {"openms", "sage", "happy"}
        assert name in names, f"Invalid score type: {name}, allowed values are: {names}"
        self.__py_ptr = psc.PyScoreType(name)

    @classmethod
    def from_py_ptr(cls, py_ptr: psc.PyScoreType):
        instance = cls.__new__(cls)
        instance.__py_ptr = py_ptr
        return instance

    def get_py_ptr(self) -> psc.PyScoreType:
        return self.__py_ptr

    def __repr__(self):
        return f"ScoreType({self.__py_ptr.to_str()})"


class PeptideSpectrumMatch:
    def __init__(self,
                 spec_idx: str,
                 peptide_idx: int,
                 proteins: List[str],
                 decoy: bool,
                 hyper_score: float,
                 rank: int,
                 mono_mass_observed: Union[None, float],
                 isotope_error: Union[None, int],
                 average_ppm: Union[None, float],
                 delta_next: Union[None, float],
                 delta_best: Union[None, float],
                 matched_peaks: Union[None, int],
                 longest_b: Union[None, int],
                 longest_y: Union[None, int],
                 longest_y_pct: Union[None, float],
                 missed_cleavages: Union[None, int],
                 matched_intensity_pct: Union[None, float],
                 scored_candidates: Union[None, int],
                 poisson: Union[None, float],
                 sequence: Union[None, str],
                 charge: Union[None, int],
                 retention_time_observed: Union[None, float],
                 retention_time_predicted: Union[None, float],
                 inverse_mobility_observed: Union[None, float],
                 inverse_mobility_predicted: Union[None, float],
                 intensity_ms1: Union[None, float],
                 intensity_ms2: Union[None, float],
                 q_value: Union[None, float],
                 collision_energy: Union[None, float],
                 collision_energy_calibrated: Union[None, float],
                 fragments: Union[None, 'Fragments'] = None,
                 re_score: Union[None, float] = None,
                 cosine_similarity: Union[None, float] = None,
                 file_name: Union[None, str] = None,
                 fragment_charges: Union[None, int] = None,
                 fragment_ion_types: Union[None, List[str]] = None,
                 fragment_ordinals: Union[None, List[int]] = None,
                 fragment_intensities: Union[None, List[float]] = None,
                 fragment_mz_calculated: Union[None, List[float]] = None,
                 fragment_mz_experimental: Union[None, List[float]] = None,
                 mz_calibration_ppm: Union[None, float] = None,
                 projected_rt: Union[None, float] = None,
                 beta_score: Union[None, float] = None,
                 ):
        self.__py_ptr = psc.PyPeptideSpectrumMatch(
            spec_idx, peptide_idx, proteins, decoy, hyper_score, rank, mono_mass_observed,
            isotope_error, average_ppm, delta_next, delta_best, matched_peaks, longest_b, longest_y,
            longest_y_pct, missed_cleavages, matched_intensity_pct, scored_candidates, poisson,
            sequence, charge,
            retention_time_observed, retention_time_predicted, inverse_mobility_observed, inverse_mobility_predicted,
            intensity_ms1, intensity_ms2, q_value, collision_energy, collision_energy_calibrated, fragments.get_py_ptr(),
            re_score, cosine_similarity, file_name,
            fragment_charges, fragment_ion_types, fragment_ordinals, fragment_intensities, fragment_mz_calculated,
            fragment_mz_experimental, mz_calibration_ppm, projected_rt, beta_score
        )

    @property
    def spec_idx(self):
        return self.__py_ptr.spec_idx

    @property
    def peptide_idx(self):
        return self.__py_ptr.peptide_idx

    @property
    def proteins(self):
        return self.__py_ptr.proteins

    @property
    def decoy(self):
        return self.__py_ptr.decoy

    @property
    def hyper_score(self):
        return self.__py_ptr.hyper_score

    @hyper_score.setter
    def hyper_score(self, value):
        self.__py_ptr.hyper_score = value

    @property
    def re_score(self):
        return self.__py_ptr.re_score

    @re_score.setter
    def re_score(self, value):
        self.__py_ptr.re_score = value

    @property
    def rank(self):
        return self.__py_ptr.rank

    @property
    def charge(self):
        return self.__py_ptr.charge

    @property
    def mono_mass_observed(self):
        return self.__py_ptr.mono_mass_observed

    @property
    def isotope_error(self):
        return self.__py_ptr.isotope_error

    @property
    def average_ppm(self):
        return self.__py_ptr.average_ppm

    @property
    def delta_next(self):
        return self.__py_ptr.delta_next

    @property
    def delta_best(self):
        return self.__py_ptr.delta_best

    @property
    def matched_peaks(self):
        return self.__py_ptr.matched_peaks

    @property
    def longest_b(self):
        return self.__py_ptr.longest_b

    @property
    def longest_y(self):
        return self.__py_ptr.longest_y

    @property
    def longest_y_pct(self):
        return self.__py_ptr.longest_y_pct

    @property
    def missed_cleavages(self):
        return self.__py_ptr.missed_cleavages

    @property
    def matched_intensity_pct(self):
        return self.__py_ptr.matched_intensity_pct

    @property
    def scored_candidates(self):
        return self.__py_ptr.scored_candidates

    @property
    def poisson(self):
        return self.__py_ptr.poisson

    @property
    def sequence(self):
        return self.__py_ptr.peptide_sequence

    @property
    def mono_mz_calculated(self):
        return self.__py_ptr.mono_mz_calculated

    @property
    def mono_mass_calculated(self):
        return self.__py_ptr.mono_mass_calculated

    @property
    def retention_time_observed(self):
        return self.__py_ptr.retention_time_observed

    @property
    def retention_time_predicted(self):
        return self.__py_ptr.retention_time_predicted

    @retention_time_predicted.setter
    def retention_time_predicted(self, value):
        self.__py_ptr.retention_time_predicted = value

    @property
    def inverse_mobility_observed(self):
        return self.__py_ptr.inverse_mobility_observed

    @property
    def inverse_mobility_predicted(self):
        return self.__py_ptr.inverse_mobility_predicted

    @inverse_mobility_predicted.setter
    def inverse_mobility_predicted(self, value):
        self.__py_ptr.inverse_mobility_predicted = value

    @property
    def intensity_ms1(self):
        return self.__py_ptr.intensity_ms1

    @property
    def intensity_ms2(self):
        return self.__py_ptr.intensity_ms2

    @property
    def q_value(self):
        return self.__py_ptr.q_value

    @property
    def collision_energy(self):
        return self.__py_ptr.collision_energy

    @collision_energy.setter
    def collision_energy(self, value):
        self.__py_ptr.collision_energy = value

    @property
    def collision_energy_calibrated(self):
        return self.__py_ptr.collision_energy_calibrated

    @collision_energy_calibrated.setter
    def collision_energy_calibrated(self, value):
        self.__py_ptr.collision_energy_calibrated = value

    @property
    def cosine_similarity(self):
        return self.__py_ptr.cosine_similarity

    @property
    def file_name(self):
        return self.__py_ptr.file_name

    @file_name.setter
    def file_name(self, value):
        self.__py_ptr.file_name = value

    @property
    def fragments_observed(self) -> Union[None, 'Fragments']:
        return Fragments.from_py_fragments(self.__py_ptr.fragments_observed)

    @property
    def fragments_predicted(self) -> Union[None, 'Fragments']:
        if self.__py_ptr.fragments_predicted is None:
            return None
        return Fragments.from_py_fragments(self.__py_ptr.fragments_predicted)

    @property
    def mz_calibration_ppm(self):
        return self.__py_ptr.mz_calibration_ppm

    @mz_calibration_ppm.setter
    def mz_calibration_ppm(self, value):
        self.__py_ptr.mz_calibration_ppm = value

    @property
    def projected_rt(self):
        return self.__py_ptr.projected_rt

    @projected_rt.setter
    def projected_rt(self, value):
        self.__py_ptr.projected_rt = value

    @property
    def beta_score(self):
        return self.__py_ptr.beta_score

    @beta_score.setter
    def beta_score(self, value):
        self.__py_ptr.beta_score = value

    @classmethod
    def from_py_ptr(cls, py_ptr: psc.PyPeptideSpectrumMatch):
        instance = cls.__new__(cls)
        instance.__py_ptr = py_ptr
        return instance

    @staticmethod
    def from_json(json_str: str) -> 'PeptideSpectrumMatch':
        psm = psc.psm_from_json(json_str)
        return PeptideSpectrumMatch.from_py_ptr(psm)

    def get_py_ptr(self) -> psc.PyPeptideSpectrumMatch:
        return self.__py_ptr

    def associate_fragment_ions_with_prosit_predicted_intensities(self, flat_intensities: List[float]):
        self.__py_ptr.associate_fragment_ions_with_prosit_predicted_intensities(flat_intensities)

    def match_observed_predicted_intensities(self) -> Tuple[Optional['Fragments'], Optional[List[float]]]:
        maybe_fragment, flat_intensities = self.__py_ptr.match_observed_predicted_intensities()
        if maybe_fragment is not None:
            return Fragments.from_py_fragments(maybe_fragment), flat_intensities
        return None, None

    def to_json(self) -> str:
        return self.__py_ptr.to_json()

    def __repr__(self):
       return (f"PeptideSpectrumMatch(spec_idx: {self.spec_idx}, match_idx: {self.peptide_idx}, "
               f"proteins: {self.proteins}, decoy: {self.decoy}, hyper_score: {self.hyper_score}, "
               f"rank: {self.rank}, mono_mass_observed: {self.mono_mass_observed}, "
               f"isotope_error: {self.isotope_error}, average_ppm: {self.average_ppm}, "
               f"delta_next: {self.delta_next}, delta_best: {self.delta_best}, "
               f"matched_peaks: {self.matched_peaks}, longest_b: {self.longest_b}, "
               f"longest_y: {self.longest_y}, longest_y_pct: {self.longest_y_pct}, "
               f"missed_cleavages: {self.missed_cleavages}, matched_intensity_pct: {self.matched_intensity_pct}, "
               f"scored_candidates: {self.scored_candidates}, poisson: {self.poisson}, "
               f"sequence: {self.sequence}, charge: {self.charge}, "
               f"retention_time_observed: {self.retention_time_observed}, "
               f"retention_time_predicted: {self.retention_time_predicted}, "
               f"inverse_mobility_observed: {self.inverse_mobility_observed}, "
               f"inverse_mobility_predicted: {self.inverse_mobility_predicted}, "
               f"intensity_ms1: {self.intensity_ms1}, intensity_ms2: {self.intensity_ms2}, "
               f"q_value: {self.q_value}, collision_energy: {self.collision_energy}, "
               f"collision_energy_calibrated: {self.collision_energy_calibrated}, "
               f"fragments_observed: {self.fragments_observed if self.fragments_predicted is not None else None}, "
               f"fragments_predicted: {self.fragments_predicted if self.fragments_observed is not None else None}, "
               f"re_score: {self.re_score if self.re_score is not None else None}, "
               f"cosine_similarity: {self.cosine_similarity if self.cosine_similarity is not None else None}, "
               f"file_name: {self.file_name if self.file_name is not None else None}, "
               f"mz_calibration_ppm: {self.mz_calibration_ppm if self.mz_calibration_ppm is not None else None}, "
               f"projected_rt: {self.projected_rt if self.projected_rt is not None else None}, "
               f"beta_score: {self.beta_score if self.beta_score is not None else None})"
               )


class Fragments:
    def __init__(self,
                charges: List[int],
                ion_types: List[IonType],
                fragment_ordinals: List[int],
                intensities: List[float],
                mz_calculated: List[float],
                mz_experimental: List[float]
                ):

        kinds = [x.get_py_ptr() for x in ion_types]
        self.__fragments_ptr = psc.PyFragments(charges, kinds, fragment_ordinals,
                                               intensities, mz_calculated, mz_experimental)

    @classmethod
    def from_py_fragments(cls, fragments: psc.PyFragments):
        instance = cls.__new__(cls)
        instance.__fragments_ptr = fragments
        return instance

    @property
    def charges(self) -> List[int]:
        return self.__fragments_ptr.charges

    @property
    def ion_types(self) -> List[IonType]:
        return [IonType.from_py_kind(x) for x in self.__fragments_ptr.kinds]

    @property
    def fragment_ordinals(self) -> List[int]:
        return self.__fragments_ptr.fragment_ordinals

    @property
    def intensities(self) -> List[float]:
        return self.__fragments_ptr.intensities

    @property
    def mz_calculated(self) -> List[float]:
        return self.__fragments_ptr.mz_calculated

    @property
    def mz_experimental(self) -> List[float]:
        return self.__fragments_ptr.mz_experimental

    def __repr__(self):
        return (f"Fragments(charges: {self.charges}, "
                f"ion_types: {self.ion_types}, "
                f"fragment_ordinals: {self.fragment_ordinals}, "
                f"intensities: {self.intensities}, "
                f"mz_calculated: {self.mz_calculated}, "
                f"mz_experimental: {self.mz_experimental})")

    def get_py_ptr(self):
        return self.__fragments_ptr


class Scorer:

    def __init__(
            self,
            precursor_tolerance: Tolerance = Tolerance(da=(-5, 5)),
            fragment_tolerance: Tolerance = Tolerance(ppm=(-10, 10)),
            min_matched_peaks: int = 6,
            min_isotope_err: int = -1,
            max_isotope_err: int = 3,
            min_precursor_charge: int = 2,
            max_precursor_charge: int = 4,
            min_fragment_mass: float = 150,
            max_fragment_mass: float = 2000,
            chimera: bool = False,
            report_psms: int = 1,
            wide_window: bool = False,
            annotate_matches: bool = False,
            score_type: ScoreType = ScoreType("openms"),
            max_fragment_charge: Optional[int] = 1,
    ):
        """Scorer class

        Args:
            precursor_tolerance (Tolerance, optional): The precursor tolerance. Defaults to Tolerance(da=(-5, 5)).
            fragment_tolerance (Tolerance, optional): The fragment tolerance. Defaults to Tolerance(ppm=(-10, 10)).
            min_matched_peaks (int, optional): The minimum number of matched peaks. Defaults to 6.
            min_isotope_err (int, optional): The minimum isotope error. Defaults to -1.
            max_isotope_err (int, optional): The maximum isotope error. Defaults to 3.
            min_precursor_charge (int, optional): The minimum precursor charge. Defaults to 2.
            max_precursor_charge (int, optional): The maximum precursor charge. Defaults to 4.
            min_fragment_mass (float, optional): The minimum fragment mass. Defaults to 150.
            max_fragment_mass (float, optional): The maximum fragment mass. Defaults to 2000.
            chimera (bool, optional): Should chimera be used. Defaults to False.
            report_psms (int, optional): The number of PSMs to report. Defaults to 1.
            wide_window (bool, optional): Should wide window be used. Defaults to False.
            annotate_matches (bool, optional): Should matches be annotated. Defaults to False.
            score_type (ScoreType, optional): The score type. Defaults to ScoreType("openms").
            max_fragment_charge (Optional[int], optional): The maximum fragment charge. Defaults to 1.
        """
        self.__scorer_ptr = psc.PyScorer(precursor_tolerance.get_py_ptr(),
                                         fragment_tolerance.get_py_ptr(),
                                         min_matched_peaks,
                                         min_isotope_err, max_isotope_err, min_precursor_charge,
                                         max_precursor_charge, min_fragment_mass, max_fragment_mass,
                                         chimera, report_psms, wide_window, annotate_matches, max_fragment_charge,
                                         score_type.get_py_ptr())

    @classmethod
    def from_py_scorer(cls, scorer: psc.PyScorer):
        instance = cls.__new__(cls)
        instance.__scorer_ptr = scorer
        return instance

    @property
    def precursor_tolerance(self) -> Tolerance:
        return Tolerance.from_py_tolerance(self.__scorer_ptr.precursor_tolerance)

    @property
    def fragment_tolerance(self) -> Tolerance:
        return Tolerance.from_py_tolerance(self.__scorer_ptr.fragment_tolerance)

    @property
    def min_matched_peaks(self) -> int:
        return self.__scorer_ptr.min_matched_peaks

    @property
    def min_isotope_err(self) -> int:
        return self.__scorer_ptr.min_isotope_err

    @property
    def max_isotope_err(self) -> int:
        return self.__scorer_ptr.max_isotope_err

    @property
    def min_precursor_charge(self) -> int:
        return self.__scorer_ptr.min_precursor_charge

    @property
    def max_precursor_charge(self) -> int:
        return self.__scorer_ptr.max_precursor_charge

    @property
    def min_fragment_mass(self) -> float:
        return self.__scorer_ptr.min_fragment_mass

    @property
    def max_fragment_mass(self) -> float:
        return self.__scorer_ptr.max_fragment_mass

    @property
    def chimera(self) -> bool:
        return self.__scorer_ptr.chimera

    @property
    def report_psms(self) -> int:
        return self.__scorer_ptr.report_psms

    @property
    def wide_window(self) -> bool:
        return self.__scorer_ptr.wide_window

    @property
    def max_fragment_charge(self) -> Optional[int]:
        if self.__scorer_ptr.max_fragment_charge is None:
            return None
        else:
            return self.__scorer_ptr.max_fragment_charge

    def __repr__(self):
        return (f"Scorer({self.precursor_tolerance}, {self.fragment_tolerance}, {self.min_matched_peaks}, "
                f"{self.min_isotope_err}, {self.max_isotope_err}, {self.min_precursor_charge}, "
                f"{self.max_precursor_charge}, {self.min_fragment_mass}, {self.max_fragment_mass}, "
                f"{self.chimera}, {self.report_psms}, {self.wide_window}, {self.max_fragment_charge})")

    def score(self, db: IndexedDatabase, spectrum: ProcessedSpectrum) -> List['Feature']:
        return [Feature.from_py_feature(f) for f in self.__scorer_ptr.score(db.get_py_ptr(), spectrum.get_py_ptr())]

    def score_collection_top_n(self, db: IndexedDatabase,
                               spectrum_collection: List[ProcessedSpectrum], num_threads: int = 4) -> List[
        List['Feature']]:
        scores = self.__scorer_ptr.score_collection(db.get_py_ptr(),
                                                    [spec.get_py_ptr() for spec in spectrum_collection], num_threads)
        return [[Feature.from_py_feature(f) for f in score] for score in scores]

    def score_collection(self, db: IndexedDatabase, spectrum_collection: List[Optional[ProcessedSpectrum]],
                         num_threads: int = 4) -> List['Feature']:
        scores = self.score_collection_top_n(db, spectrum_collection, num_threads)

        result = []

        for score in scores:
            if len(score) > 0:
                result.append(score[0])
            else:
                result.append(None)

        return result

    def score_collection_psm(self, db: IndexedDatabase, spectrum_collection: List[Optional[ProcessedSpectrum]],
                             num_threads: int = 4) -> Dict[str, List[PeptideSpectrumMatch]]:

        py_psms = self.__scorer_ptr.score_collection_to_psm_collection(db.get_py_ptr(),
                                                                       [spec.get_py_ptr() for spec in
                                                                        spectrum_collection],
                                                                       num_threads)

        ret_dict = {}
        for key, values in py_psms.items():
            ret_dict[key] = [PeptideSpectrumMatch.from_py_ptr(psm) for psm in values]

        return ret_dict

    def score_collection_psm_pandas(self, db: IndexedDatabase, spectrum_collection: List[Optional[ProcessedSpectrum]],
                                    num_threads: int = 4) -> pd.DataFrame:

        py_psms = self.__scorer_ptr.score_collection_to_psm_collection(db.get_py_ptr(),
                                                                       [spec.get_py_ptr() for spec in
                                                                        spectrum_collection],
                                                                       num_threads)
        row_list = []
        for match in py_psms:
            row_list.append({
                "spec_idx": match.spec_idx,
                "match_idx": match.peptide_idx,
                "proteins": match.proteins,
                "decoy": match.decoy,
                "score": match.hyper_score,
                "rank": match.rank,
                "mono_mz_calculated": match.mono_mz_calculated,
                "mono_mass_observed": match.mono_mass_observed,
                "mono_mass_calculated": match.mono_mass_calculated,
                "sequence": match.peptide_sequence,
                "charge": match.charge,
                "retention_time_observed": match.retention_time_observed,
                "retention_time_predicted": match.retention_time_predicted,
                "inverse_mobility_observed": match.inverse_mobility_observed,
                "inverse_mobility_predicted": match.inverse_mobility_predicted,
                "intensity_ms1": match.intensity_ms1,
                "intensity_ms2": match.intensity_ms2,
                "q_value": match.q_value,
                "collision_energy": match.collision_energy,
            })

        return pd.DataFrame(row_list)

    def _score_chimera_fast(self, db: IndexedDatabase, spectrum: ProcessedSpectrum) -> List['Feature']:
        return [Feature.from_py_feature(f) for f in
                self.__scorer_ptr.score_chimera_fast(db.get_py_ptr(), spectrum.get_py_ptr())]

    def _score_standard(self, db: IndexedDatabase, spectrum: ProcessedSpectrum) -> List['Feature']:
        return [Feature.from_py_feature(f) for f in
                self.__scorer_ptr.score_standard(db.get_py_ptr(), spectrum.get_py_ptr())]


class Feature:
    def __init__(self, peptide_idx: PeptideIx, psm_id: int, peptide_len: int, spec_id: str, file_id: int,
                 rank: int, label: int, expmass: float, calcmass: float, charge: int, delta_mass: float,
                 isotope_error: float, average_ppm: float, hyperscore: float, delta_next: float,
                 delta_best: float, matched_peaks: int, longest_b: int, longest_y: int,
                 longest_y_pct: float, missed_cleavages: int, matched_intensity_pct: float,
                 scored_candidates: int, poisson: float, discriminant_score: float,
                 posterior_error: float, spectrum_q: float, peptide_q: float, protein_q: float,
                 ms2_intensity: float,
                 fragments: Optional[float] = None,
                 rt: Optional[float] = None,
                 aligned_rt: Optional[float] = None,
                 predicted_rt: Optional[float] = None,
                 delta_rt_model: Optional[float] = None,
                 ims: Optional[float] = None,
                 predicted_ims: Optional[float] = None,
                 delta_ims_model: Optional[float] = None, ):
        """Feature class

        Args:
            peptide_idx (PeptideIx): The peptide index
            psm_id (int): The PSM id
            peptide_len (int): The peptide length
            spec_id (str): The spectrum id
            file_id (int): The file id
            rank (int): The rank
            label (int): The label
            expmass (float): The experimental mass
            calcmass (float): The calculated mass
            charge (int): The charge
            rt (float): The retention time
            aligned_rt (float): The aligned retention time
            predicted_rt (float): The predicted retention time
            delta_rt_model (float): The delta retention time model
            delta_mass (float): The delta mass
            isotope_error (float): The isotope error
            average_ppm (float): The average ppm
            hyperscore (float): The hyperscore
            delta_next (float): The delta next
            delta_best (float): The delta best
            matched_peaks (int): The number of matched peaks
            longest_b (int): The longest b ion
            longest_y (int): The longest y ion
            longest_y_pct (float): The longest y ion percentage
            missed_cleavages (int): The number of missed cleavages
            matched_intensity_pct (float): The matched intensity percentage
            scored_candidates (int): The number of scored candidates
            poisson (float): The poisson
            discriminant_score (float): The discriminant score
            posterior_error (float): The posterior error
            spectrum_q (float): The spectrum q
            peptide_q (float): The peptide q
            protein_q (float): The protein q
            ms2_intensity (float): The MS2 intensity
        """

        self.__feature_ptr = psc.PyFeature(peptide_idx, psm_id, peptide_len, spec_id, file_id, rank, label,
                                           expmass, calcmass, charge, delta_mass, isotope_error, average_ppm,
                                           hyperscore, delta_next, delta_best, matched_peaks,
                                           longest_b, longest_y, longest_y_pct, missed_cleavages,
                                           matched_intensity_pct, scored_candidates, poisson,
                                           discriminant_score, posterior_error, spectrum_q,
                                           peptide_q, protein_q, ms2_intensity,
                                           fragments.get_py_ptr() if fragments is not None else None,
                                           rt,
                                           aligned_rt,
                                           predicted_rt,
                                           delta_rt_model,
                                           ims,
                                           predicted_ims,
                                           delta_ims_model, )

    @classmethod
    def from_py_feature(cls, feature: psc.PyFeature):
        instance = cls.__new__(cls)
        instance.__feature_ptr = feature
        return instance

    @property
    def peptide_idx(self) -> PeptideIx:
        return PeptideIx.from_py_peptide_ix(self.__feature_ptr.peptide_idx)

    @property
    def peptide_len(self) -> int:
        return self.__feature_ptr.peptide_len

    @property
    def spec_id(self) -> str:
        return self.__feature_ptr.spec_id

    @property
    def psm_id(self) -> int:
        return self.__feature_ptr.psm_id

    @property
    def file_id(self) -> int:
        return self.__feature_ptr.file_id

    @property
    def rank(self) -> int:
        return self.__feature_ptr.rank

    @property
    def label(self) -> int:
        return self.__feature_ptr.label

    @property
    def expmass(self) -> float:
        return self.__feature_ptr.expmass

    @property
    def calcmass(self) -> float:
        return self.__feature_ptr.calcmass

    @property
    def charge(self) -> int:
        return self.__feature_ptr.charge

    @property
    def rt(self) -> float:
        return self.__feature_ptr.rt

    @property
    def aligned_rt(self) -> float:
        return self.__feature_ptr.aligned_rt

    @property
    def predicted_rt(self) -> float:
        return self.__feature_ptr.predicted_rt

    @property
    def delta_rt_model(self) -> float:
        return self.__feature_ptr.delta_rt_model

    @property
    def delta_mass(self) -> float:
        return self.__feature_ptr.delta_mass

    @property
    def isotope_error(self) -> float:
        return self.__feature_ptr.isotope_error

    @property
    def average_ppm(self) -> float:
        return self.__feature_ptr.average_ppm

    @property
    def hyperscore(self) -> float:
        return self.__feature_ptr.hyperscore

    @property
    def delta_next(self) -> float:
        return self.__feature_ptr.delta_next

    @property
    def delta_best(self) -> float:
        return self.__feature_ptr.delta_best

    @property
    def matched_peaks(self) -> int:
        return self.__feature_ptr.matched_peaks

    @property
    def longest_b(self) -> int:
        return self.__feature_ptr.longest_b

    @property
    def longest_y(self) -> int:
        return self.__feature_ptr.longest_y

    @property
    def longest_y_pct(self) -> float:
        return self.__feature_ptr.longest_y_pct

    @property
    def missed_cleavages(self) -> int:
        return self.__feature_ptr.missed_cleavages

    @property
    def matched_intensity_pct(self) -> float:
        return self.__feature_ptr.matched_intensity_pct

    @property
    def scored_candidates(self) -> int:
        return self.__feature_ptr.scored_candidates

    @property
    def poisson(self) -> float:
        return self.__feature_ptr.poisson

    @property
    def discriminant_score(self) -> float:
        return self.__feature_ptr.discriminant_score

    @property
    def posterior_error(self) -> float:
        return self.__feature_ptr.posterior_error

    @property
    def spectrum_q(self) -> float:
        return self.__feature_ptr.spectrum_q

    @property
    def peptide_q(self) -> float:
        return self.__feature_ptr.peptide_q

    @property
    def protein_q(self) -> float:
        return self.__feature_ptr.protein_q

    @property
    def ms2_intensity(self) -> float:
        return self.__feature_ptr.ms2_intensity

    @property
    def fragments(self) -> Optional[Fragments]:
        if self.__feature_ptr.fragments is None:
            return None
        else:
            return Fragments.from_py_fragments(self.__feature_ptr.fragments)

    @property
    def ims(self) -> Optional[float]:
        return self.__feature_ptr.ims

    @property
    def predicted_ims(self) -> Optional[float]:
        return self.__feature_ptr.predicted_ims

    @property
    def delta_ims_model(self) -> Optional[float]:
        return self.__feature_ptr.delta_ims_model

    def __repr__(self):
        return (f"Feature("
                f"idx: {self.peptide_idx}, "
                f"psm_id: {self.psm_id}, "
                f"peptide_len: {self.peptide_len}, "
                f"spec_id: {self.spec_id}, "
                f"file_id: {self.file_id}, "
                f"rank: {self.rank}, "
                f"label: {self.label}, "
                f"exp. mass: {self.expmass}, "
                f"cal. mass: {self.calcmass}, "
                f"charge: {self.charge}, "
                f"retention time: {self.rt}, "
                f"aligned rt: {self.aligned_rt}, "
                f"predicted rt: {self.predicted_rt}, "
                f"delta rt model: {self.delta_rt_model}, "
                f"delta mass: {self.delta_mass}, "
                f"isotope error: {self.isotope_error}, "
                f"average ppm: {self.average_ppm}, "
                f"hyperscore: {self.hyperscore}, "
                f"delta_next: {self.delta_next}, "
                f"delta_best: {self.delta_best}, "
                f"matched peaks: {self.matched_peaks}, "
                f"longest b: {self.longest_b},"
                f"longest y: {self.longest_y}, "
                f"longest y pct: {self.longest_y_pct}, "
                f"missed cleavages: {self.missed_cleavages}, "
                f"matched intensity pct: {self.matched_intensity_pct}, "
                f"scored candidates: {self.scored_candidates}, "
                f"poisson: {self.poisson}, "
                f"discriminant score: {self.discriminant_score}, "
                f"posterior error: {self.posterior_error}, "
                f"spectrum q: {self.spectrum_q}, "
                f"peptide q: {self.peptide_q}, "
                f"protein q: {self.protein_q}, "
                f"ms2 intensity: {self.ms2_intensity}), "
                f"fragments: {self.fragments})")

    def get_py_ptr(self):
        return self.__feature_ptr


def associate_fragment_ions_with_prosit_predicted_intensities(
        psms: List[PeptideSpectrumMatch],
        flat_intensities: List[List[float]], num_threads: int = 16) -> List['PeptideSpectrumMatch']:
    """Associate fragment ions with prosit predicted intensities in parallel

    Args:
        psms (List[PeptideSpectrumMatch]): The peptide spectrum matches
        flat_intensities (List[List[float]]): The flat intensities
        num_threads (int, optional): The number of threads. Defaults to 16.

    Returns:
        List[PeptideSpectrumMatch]: The peptide spectrum matches
    """
    result = psc.associate_fragment_ions_with_prosit_predicted_intensities_par(
        [psm.get_py_ptr() for psm in psms], flat_intensities, num_threads
    )
    return [PeptideSpectrumMatch.from_py_ptr(f) for f in result]


def associate_fragment_ions_with_prosit_predicted_intensities_pandas(
        psms: List[PeptideSpectrumMatch],
        flat_intensities: List[List[float]],
        num_threads: int = 16,
) -> pd.DataFrame:
    """Associate fragment ions with prosit predicted intensities in parallel

    Args:
        psms (List[PeptideSpectrumMatch]): The peptide spectrum matches
        flat_intensities (List[List[float]]): The flat intensities
        num_threads (int, optional): The number of threads. Defaults to 16.

    Returns:
        pd.DataFrame: The pandas dataframe
    """
    result = psc.associate_fragment_ions_with_prosit_predicted_intensities_par(
        [psm.get_py_ptr() for psm in psms], flat_intensities, num_threads
    )
    row_list = []
    for match in result:
        row_list.append({
            "spec_idx": match.spec_idx,
            "match_idx": match.peptide_idx,
            "proteins": match.proteins,
            "decoy": match.decoy,
            "score": match.hyper_score,
            "rank": match.rank,
            "mono_mz_calculated": match.mono_mz_calculated,
            "mono_mass_observed": match.mono_mass_observed,
            "mono_mass_calculated": match.mono_mass_calculated,
            "sequence": match.peptide_sequence,
            "charge": match.charge,
            "retention_time_observed": match.retention_time_observed,
            "retention_time_predicted": match.retention_time_predicted,
            "inverse_mobility_observed": match.inverse_mobility_observed,
            "inverse_mobility_predicted": match.inverse_mobility_predicted,
            "intensity_ms1": match.intensity_ms1,
            "intensity_ms2": match.intensity_ms2,
            "q_value": match.q_value,
            "collision_energy": match.collision_energy,
            "cosine_similarity": match.cosine_similarity,
        })
    return pd.DataFrame(row_list)


def json_bin_to_psms(json_bin: bytes) -> List[PeptideSpectrumMatch]:
    """ Convert a binary JSON string to a list of PeptideSpectrumMatch objects.

    Args:
        json_bin: a binary JSON string

    Returns:
        a list of PeptideSpectrumMatch objects
    """
    return [PeptideSpectrumMatch.from_py_ptr(json_str) for json_str in psc_utils.json_bin_to_psms(json_bin)]


def psms_to_json(psms, num_threads: int = 4) -> List[str]:
    """ Convert a list of PeptideSpectrumMatch objects to a JSON string.

    Args:
        psms: a list of PeptideSpectrumMatch objects
        num_threads: the number of threads to use

    Returns:
        a JSON string
    """
    return psc_utils.psms_to_json([psm.get_py_ptr() for psm in psms], num_threads)


def psms_to_json_bin(psms) -> bytes:
    """ Convert a list of PeptideSpectrumMatch objects to a binary JSON string.

    Args:
        psms: a list of PeptideSpectrumMatch objects

    Returns:
        a binary JSON string
    """
    return psc_utils.psms_to_json_bin([psm.get_py_ptr() for psm in psms])


def merge_psm_dicts(left_psms: Dict[str, List[PeptideSpectrumMatch]],
                    right_psms: Dict[str, List[PeptideSpectrumMatch]],
                    max_hits: int = 5) -> Dict[str, List[PeptideSpectrumMatch]]:
    """ Merge two dictionaries of peptide spectrum matches.

    Args:
        left_psms: the left dictionary of peptide spectrum matches
        right_psms: the right dictionary of peptide spectrum matches
        max_hits: the maximum number of hits

    Returns:
        a dictionary of peptide spectrum matches
    """
    left_map = {key: [psm.get_py_ptr() for psm in value] for key, value in left_psms.items()}
    right_map = {key: [psm.get_py_ptr() for psm in value] for key, value in right_psms.items()}
    result = psc.merge_psm_maps(left_map, right_map, max_hits)
    return {key: [PeptideSpectrumMatch.from_py_ptr(psm) for psm in value] for key, value in result.items()}


def prosit_intensities_to_fragments(
        flat_intensities: List[float],
) -> Fragments:
    """ Convert a list of intensities to a Fragments object.

    Args:
        flat_intensities: a list of intensities

    Returns:
        a Fragments object
    """
    return Fragments.from_py_fragments(psc.prosit_intensities_to_py_fragments(flat_intensities))


def prosit_intensities_to_fragments_par(
        flat_intensities: List[float],
        num_threads: int = 16,
) -> List[Fragments]:
    """ Convert a list of intensities to a Fragments object in parallel.

    Args:
        flat_intensities: a list of intensities
        num_threads: the number of threads

    Returns:
        a Fragments List object
    """

    return [Fragments.from_py_fragments(f) for f in psc.prosit_intensities_to_py_fragments_par(flat_intensities, num_threads)]



import sagepy_connector

from sagepy.core.peptide import Peptide

psc = sagepy_connector.py_ion_series


class IonType:
    def __init__(self, ion_type: str):
        """IonType class

        Args:
            ion_type (str): The ion type, allowed values are: a, b, c, x, y, z
        """
        try:
            self.__ion_type_ptr = psc.PyKind(ion_type)
        except ValueError:
            raise ValueError("Invalid ion type, allowed values are: a, b, c, x, y, z")

    @classmethod
    def from_py_kind(cls, kind: psc.PyKind):
        instance = cls.__new__(cls)
        instance.__ion_type_ptr = kind
        return instance

    @classmethod
    def y(cls):
        return cls.from_py_kind(psc.PyKind('y'))

    @classmethod
    def b(cls):
        return cls.from_py_kind(psc.PyKind('b'))

    def __repr__(self):
        return f"IonType({self.__ion_type_ptr.kind_as_string()})"

    def __hash__(self):
        return hash(self.__ion_type_ptr.kind_as_string())

    def __eq__(self, other):
        if not isinstance(other, IonType):
            return False
        return self.__ion_type_ptr.kind_as_string() == other.__ion_type_ptr.kind_as_string()

    def get_py_ptr(self):
        return self.__ion_type_ptr

    def to_string(self) -> str:
        return self.__ion_type_ptr.kind_as_string()


class Ion:
    """Ion class

    Args:
        ion_type (IonType): The ion type, e.g. b, y
        mass (float): The mass of the ion
    """
    def __init__(self, ion_type: IonType, mass: float):
        self.__ion_ptr = psc.PyIon(ion_type.get_py_ptr(), mass)

    @classmethod
    def from_py_ion(cls, ion: psc.PyIon):
        instance = cls.__new__(cls)
        instance.__ion_ptr = ion
        return instance

    @property
    def ion_type(self):
        return IonType.from_py_kind(self.__ion_ptr.kind)

    @property
    def mono_isotopic_mass(self):
        return self.__ion_ptr.monoisotopic_mass

    def __repr__(self):
        return f"Ion({self.ion_type}, {self.mono_isotopic_mass})"


class IonSeries:
    """IonSeries class

    Args:
        peptide (Peptide): The peptide
        ion_type (IonType): The ion type, e.g. b, y
    """
    def __init__(self, peptide: Peptide, ion_type: IonType):
        self.__ion_series_ptr = psc.PyIonSeries(peptide.get_py_ptr(), ion_type.get_py_ptr())

    @classmethod
    def from_py_ion_series(cls, ion_series: psc.PyIonSeries):
        instance = cls.__new__(cls)
        instance.__ion_series_ptr = ion_series
        return instance

    @property
    def ion_type(self):
        return IonType.from_py_kind(self.__ion_series_ptr.kind)

    @property
    def cumulative_mass(self):
        return self.__ion_series_ptr.cumulative_mass

    @property
    def peptide(self):
        return Peptide.from_py_peptide(self.__ion_series_ptr.peptide)

    def __repr__(self):
        return f"IonSeries({self.ion_type}, {self.cumulative_mass}, {self.peptide})"

    def get_py_ptr(self):
        return self.__ion_series_ptr

    def get_ion_series(self):
        return [Ion.from_py_ion(i) for i in self.__ion_series_ptr.get_ion_series()]


from typing import Union

import numpy as np
import sagepy_connector
psc = sagepy_connector.py_enzyme


class Position:
    def __init__(self, position: Union[str, None] = None):
        """Position class representing location of peptide in protein

        Args:
            position (Union[str, None], optional): The position of the peptide in the protein. Defaults to None.
        """
        if position is not None:
            try:
                self.__position_ptr = psc.PyPosition.from_string(position)
            except ValueError:
                raise ValueError("Invalid position string, allowed values are: c_term, n_term, full, internal")
        else:
            self.__position_ptr = psc.PyPosition.from_string('internal')

    @classmethod
    def from_py_position(cls, position: psc.PyPosition):
        instance = cls.__new__(cls)
        instance.__position_ptr = position
        return instance

    @property
    def position(self):
        return self.__position_ptr.to_string

    def __repr__(self):
        return f"Position({self.position})"

    def get_py_ptr(self):
        return self.__position_ptr


class Enzyme:
    def __init__(self, cleave_pattern: str = 'KR', c_terminal: bool = True, skip_suffix: str = 'P', semi_enzymatic: bool = False):
        """Enzyme class, default enzyme is Trypsin

        Args:
            cleave_pattern (str, optional): The cleavage pattern of the enzyme. Defaults to 'KR'.
            c_terminal (bool, optional): Cleave from the C-terminal. Defaults to True.
            skip_suffix (str, optional): The skip suffix of the enzyme. Defaults to 'P'.
            semi_enzymatic (bool, optional): Is the enzyme semi enzymatic. Defaults to False.
        """
        self.__enzyme_ptr = psc.PyEnzyme(cleave_pattern, c_terminal, skip_suffix, semi_enzymatic)

    @classmethod
    def from_py_enzyme(cls, enzyme: psc.PyEnzyme):
        instance = cls.__new__(cls)
        instance.__enzyme_ptr = enzyme
        return instance

    @property
    def c_terminal(self):
        if self.__enzyme_ptr is None:
            return None
        return self.__enzyme_ptr.c_terminal

    @property
    def skip_suffix(self):
        if self.__enzyme_ptr is None:
            return None
        return self.__enzyme_ptr.skip_suffix

    def cleavage_sites(self, sequence: str):
        if self.__enzyme_ptr is None:
            return None
        return self.__enzyme_ptr.cleavage_sites(sequence)

    def cleave(self, sequence: str, min_length: int = 1, max_length: int = np.inf):
        if self.__enzyme_ptr is None:
            raise ValueError("Enzyme is not defined")
        cleave_sites = self.__enzyme_ptr.cleavage_sites(sequence)

        cleaved_peptides = [sequence[start:end] for start, end in cleave_sites]
        filtered_peptides = [peptide for peptide in cleaved_peptides if min_length <= len(peptide) <= max_length]

        return filtered_peptides

    def get_py_ptr(self):
        return self.__enzyme_ptr

    def __repr__(self):
        return f"Enzyme(c_terminal: {self.c_terminal}, skip_suffix: {self.skip_suffix})"


class Digest:
    def __init__(self, decoy: bool, sequence: str, protein: str, missed_cleavages: int, position: Position):
        """Digest class representing a peptide digest

        Args:
            decoy (bool): Is the digest peptide a decoy
            sequence (str): The sequence of the digest peptide
            protein (str): The protein that the digest peptide is found in
            missed_cleavages (int): The number of missed cleavages
            position (Position): The position of the digest peptide in the protein
        """
        self.__digest_ptr = psc.PyDigest(decoy, sequence, protein, missed_cleavages, position.get_py_ptr())

    @classmethod
    def from_py_digest(cls, digest: psc.PyDigest):
        instance = cls.__new__(cls)
        instance.__digest_ptr = digest
        return instance

    @property
    def decoy(self):
        return self.__digest_ptr.decoy

    @property
    def sequence(self):
        return self.__digest_ptr.sequence

    @property
    def protein(self):
        return self.__digest_ptr.protein

    @property
    def missed_cleavages(self):
        return self.__digest_ptr.missed_cleavages

    @property
    def position(self):
        return self.__digest_ptr.position

    def reverse(self):
        return self.__digest_ptr.reverse()

    def __eq__(self, other):
        return self.__digest_ptr == other.__digest_ptr

    def __hash__(self):
        return hash(self.__digest_ptr)

    def __repr__(self):
        return f"Digest(decoy: {self.decoy}, protein: {self.protein}, " \
               f"missed_cleavages: {self.missed_cleavages}, position: {self.position}, sequence: {self.sequence})"

    def get_py_ptr(self):
        return self.__digest_ptr


class EnzymeParameters:
    def __init__(self, missed_cleavages: int = 0, min_len: int = 5, max_len: int = 50, enzyme: Enzyme = None):
        """EnzymeParameters class representing the parameters of an enzyme

        Args:
            missed_cleavages (int, optional): The number of missed cleavages. Defaults to 0.
            min_len (int, optional): The minimum length of a peptide. Defaults to 5.
            max_len (int, optional): The maximum length of a peptide. Defaults to 50.
            enzyme (Enzyme, optional): The enzyme. Defaults to None.
        """
        if enzyme is None:
            self.__enzyme_parameters_ptr = psc.PyEnzymeParameters(missed_cleavages, min_len, max_len, None)
        else:
            self.__enzyme_parameters_ptr = psc.PyEnzymeParameters(missed_cleavages, min_len, max_len,
                                                                  enzyme.get_py_ptr())

    @classmethod
    def from_py_enzyme_parameters(cls, enzyme_parameters: psc.PyEnzymeParameters):
        instance = cls.__new__(cls)
        instance.__enzyme_parameters_ptr = enzyme_parameters
        return instance

    @property
    def missed_cleavages(self):
        return self.__enzyme_parameters_ptr.missed_cleavages

    @property
    def min_len(self):
        return self.__enzyme_parameters_ptr.min_len

    @property
    def max_len(self):
        return self.__enzyme_parameters_ptr.max_len

    @property
    def enzyme(self):
        return Enzyme.from_py_enzyme(self.__enzyme_parameters_ptr.enzyme)

    def __repr__(self):
        return f"EnzymeParameters(missed_cleavages: {self.missed_cleavages}, min_len: {self.min_len}, " \
               f"max_len: {self.max_len}, enzyme: {self.enzyme})"

    def cleavage_sites(self, sequence: str):
        return self.__enzyme_parameters_ptr.cleavage_sites(sequence)

    def digest(self, sequence: str, protein: str):
        return [Digest.from_py_digest(s) for s in self.__enzyme_parameters_ptr.digest(sequence, protein)]

    def get_py_ptr(self):
        return self.__enzyme_parameters_ptr


from typing import List, Tuple, Optional

import pandas as pd
import sagepy_connector

psc = sagepy_connector.py_qfdr


class TDCMethod:
    def __init__(self, method: str):
        self.methods = {"psm", "peptide_psm_only", "peptide_peptide_only", "peptide_psm_peptide"}
        assert method in self.methods, f"Invalid method: {method}, allowed values are: {self.methods}"
        self.__py_ptr = psc.PyTDCMethod(method)

    @classmethod
    def from_py_ptr(cls, py_ptr: psc.PyTDCMethod):
        instance = cls.__new__(cls)
        instance.__py_ptr = py_ptr
        return instance

    def get_py_ptr(self) -> psc.PyTDCMethod:
        return self.__py_ptr

    def __repr__(self):
        return f"TDCMethod({self.__py_ptr.to_str()})"


def target_decoy_competition(
        spectra_idx: List[str],
        match_idx: List[str],
        decoy: List[bool],
        scores: List[float],
        method: str = "peptide_psm_peptide") -> Tuple[List[str], List[str], List[bool], List[float], List[float]]:
    """ Perform target-decoy competition.

    Args:
        spectra_idx: a list of spectrum indices
        match_idx: a list of match indices
        decoy: a list of decoy flags
        scores: a list of scores
        method: the method to use, allowed values are: psm, peptide_psm_only, peptide_peptide_only, peptide_psm_peptide

    Returns:
        a tuple of spectrum indices, match indices, decoy flags, scores, and q-values
    """
    tdc_method = TDCMethod(method)
    spec_idx, match_idx, decoy, scores, q_values = psc.target_decoy_competition(
        tdc_method.get_py_ptr(), spectra_idx,
        match_idx, decoy, scores)
    return spec_idx, match_idx, decoy, scores, q_values


def target_decoy_competition_pandas(
        df: pd.DataFrame,
        method: str = "peptide_psm_peptide",
        score: Optional[str] = None,
) -> pd.DataFrame:
    """ Perform target-decoy competition on a pandas DataFrame.

    Args:
        df: a pandas DataFrame
        method: the method to use, allowed values are: psm, peptide_psm_only, peptide_peptide_only, peptide_psm_peptide
        score: the target column name

    Returns:
        a pandas DataFrame with q-values
    """
    assert 'spec_idx' in df.columns, "spec_idx column not found"
    assert 'match_idx' in df.columns, "match_idx column not found"
    assert 'decoy' in df.columns, "decoy column not found"

    if score is not None:
        assert score in df.columns, f"{score} column not found"

    else:
        assert 'score' in df.columns, "score column not found"

    score = df['score'] if score is None else df[score]

    spec_idx, match_idx, target, scores = (df['spec_idx'].tolist(),
                                           df['match_idx'].tolist(), df['decoy'].tolist(), score.tolist())

    spec_idx, match_idx, target, scores, q_values = target_decoy_competition(spec_idx,
                                                                             match_idx, target, scores, method)

    return pd.DataFrame({
        'spec_idx': spec_idx,
        'match_idx': match_idx,
        'decoy': target,
        'score': scores,
        'q_value': q_values
    }).sort_values(by=['q_value'])



from typing import Optional
from sagepy.core.database import PeptideIx
import sagepy_connector
psc = sagepy_connector.py_fdr


class CompetitionPeptideIx:
    def __init__(self, forward: float, reverse: float,
                 forward_ix: Optional[PeptideIx] = None, reverse_ix: Optional[PeptideIx] = None):

        self.__ptr = psc.PyCompetitionPeptideIx(forward, reverse, forward_ix.get_py_ptr() if forward_ix else None,
                                                reverse_ix.get_py_ptr() if reverse_ix else None)

    @classmethod
    def from_py_competition_peptide_ix(cls, competition_peptide_ix: psc.PyCompetitionPeptideIx):
        instance = cls.__new__(cls)
        instance.__ptr = competition_peptide_ix
        return instance

    def get_py_ptr(self):
        return self.__ptr

    @property
    def forward(self) -> float:
        return self.__ptr.forward

    @property
    def reverse(self) -> float:
        return self.__ptr.reverse

    @property
    def forward_ix(self) -> Optional[PeptideIx]:
        maybe_forward = self.__ptr.forward_ix
        if maybe_forward is None:
            return None
        return PeptideIx.from_py_peptide_ix(maybe_forward)

    @property
    def reverse_ix(self) -> Optional[PeptideIx]:
        maybe_reverse = self.__ptr.reverse_ix
        if maybe_reverse is None:
            return None
        return PeptideIx.from_py_peptide_ix(maybe_reverse)

    def __repr__(self):
        return (f"CompetitionPeptideIx(forward={self.forward}, reverse={self.reverse}, "
                f"forward_ix={self.forward_ix}, reverse_ix={self.reverse_ix})")



from typing import List

import sagepy_connector
psc = sagepy_connector.py_mass


def monoisotopic(aa: str) -> float:
    return psc.py_monoisotopic(aa)


class Composition:
    def __init__(self, carbon, sulfur):
        """Composition class

        Args:
            carbon (int): The number of carbon atoms
            sulfur (int): The number of sulfur atoms
        """
        self.__composition_ptr = psc.PyComposition(carbon, sulfur)

    @classmethod
    def from_py_composition(cls, composition: psc.PyComposition):
        instance = cls.__new__(cls)
        instance.__composition_ptr = composition
        return instance

    @property
    def carbon(self):
        return self.__composition_ptr.carbon

    @property
    def sulfur(self):
        return self.__composition_ptr.sulfur

    def __repr__(self):
        return f"Composition(carbon: {self.carbon}, sulfur: {self.sulfur})"

    def get_py_ptr(self):
        return self.__composition_ptr

    @staticmethod
    def sum(composition_list: List['Composition']) -> 'Composition':
        return psc.PyComposition.sum([c.get_py_ptr() for c in composition_list])

    @staticmethod
    def aa_composition(aa: str) -> 'Composition':
        return Composition.from_py_composition(psc.PyComposition.py_composition(aa))


class CONSTANTS:
    @classmethod
    @property
    def NEUTRON(cls):
        return psc.neutron()

    @classmethod
    @property
    def PROTON(cls):
        return psc.proton()

    @classmethod
    @property
    def H2O(cls):
        return psc.h2o()

    @classmethod
    @property
    def NH3(cls):
        return psc.nh3()


class Tolerance:
    def __init__(self, da: (float, float) = None, ppm: (float, float) = None):
        """Tolerance class

        Args:
            da (float, optional): The tolerance in Da. Defaults to None.
            ppm (float, optional): The tolerance in ppm. Defaults to None.
        """
        if da is not None and ppm is not None:
            raise ValueError("Only one of da or ppm can be set")
        elif da is None and ppm is None:
            raise ValueError("One of da or ppm must be set")
        else:
            self.__tolerance_ptr = psc.PyTolerance(da, ppm)

    def get_py_ptr(self):
        return self.__tolerance_ptr

    @classmethod
    def from_py_tolerance(cls, tolerance: psc.PyTolerance) -> 'Tolerance':
        instance = cls.__new__(cls)
        instance.__tolerance_ptr = tolerance
        return instance

    @property
    def da(self) -> (float, float):
        return self.__tolerance_ptr.da

    @property
    def ppm(self) -> (float, float):
        return self.__tolerance_ptr.ppm

    def bounds(self, center: float) -> (float, float):
        return self.__tolerance_ptr.bounds(center)

    def contains(self, center: float, target: float) -> bool:
        return self.__tolerance_ptr.contains(center, target)

    def __repr__(self) -> str:
        if self.da is not None:
            return f"Tolerance(da={self.da})"
        else:
            return f"Tolerance(ppm={self.ppm})"

    def __mul__(self, other) -> 'Tolerance':
        if isinstance(other, float):
            return Tolerance.from_py_tolerance(self.__tolerance_ptr * other)

        elif isinstance(other, int):
            return Tolerance.from_py_tolerance(self.__tolerance_ptr * float(other))

        else:
            raise ValueError("Tolerance can only be multiplied by a float or an int")



import numpy as np

from typing import List, Dict, Tuple, Union

import pandas as pd

from sagepy.core.mass import Tolerance
from sagepy.core.peptide import Peptide

from sagepy.core.enzyme import EnzymeParameters
import sagepy_connector

from sagepy.core.ion_series import IonType
from sagepy.core.modification import ModificationSpecificity

psc = sagepy_connector.py_database


class PeptideIx:
    def __init__(self, idx: int):
        """ PeptideIx class

        Args:
            idx (int): idx in the database
        """
        self.__peptide_ix_ptr = psc.PyPeptideIx(idx)

    @classmethod
    def from_py_peptide_ix(cls, ix: psc.PyPeptideIx) -> 'PeptideIx':
        instance = cls.__new__(cls)
        instance.__peptide_ix_ptr = ix
        return instance

    @property
    def idx(self) -> int:
        return self.__peptide_ix_ptr.idx

    def __repr__(self) -> str:
        return f"PeptideIx({self.idx})"

    def get_py_ptr(self):
        return self.__peptide_ix_ptr


class Theoretical:
    def __init__(self, idx: int, fragment_mz: float):
        """Theoretical class

        Args:
            idx (int): idx in the database
            fragment_mz (float): fragment m/z
        """
        self.__theoretical_ptr = psc.PyTheoretical(idx, fragment_mz)

    @classmethod
    def from_py_theoretical(cls, theoretical: psc.PyTheoretical) -> 'Theoretical':
        instance = cls.__new__(cls)
        instance.__theoretical_ptr = theoretical
        return instance

    @property
    def idx(self):
        return PeptideIx.from_py_peptide_ix(self.__theoretical_ptr.idx)

    @property
    def fragment_mz(self) -> float:
        return self.__theoretical_ptr.fragment_mz

    def __repr__(self) -> str:
        return f"Theoretical(idx={self.idx}, fragment_mz={self.fragment_mz})"

    def get_py_ptr(self):
        return self.__theoretical_ptr


class EnzymeBuilder:
    def __init__(self, missed_cleavages: int = None, min_len: int = None, max_len: int = None, cleave_at: str = None,
                 restrict: str = None, c_terminal: bool = None, semi_enzymatic: bool = None):
        """EnzymeBuilder class

        Args:
            missed_cleavages (int, optional): Number of missed cleavages. Defaults to None.
            min_len (int, optional): Minimum peptide length. Defaults to None.
            max_len (int, optional): Maximum peptide length. Defaults to None.
            cleave_at (str, optional): Cleavage pattern. Defaults to None.
            restrict (str, optional): Restriction pattern. Defaults to None.
            c_terminal (bool, optional): Cleavage at the C-terminal. Defaults to None.
            semi_enzymatic (bool, optional): Allow semi-enzymatic peptides. Defaults to None.
        """
        self.__enzyme_builder_ptr = psc.PyEnzymeBuilder(missed_cleavages, min_len, max_len, cleave_at, restrict,
                                                        c_terminal, semi_enzymatic)

    @staticmethod
    def default_trypsin() -> 'EnzymeBuilder':
        return EnzymeBuilder.from_py_enzyme_builder(psc.PyEnzymeBuilder.from_default_trypsin())

    @classmethod
    def from_py_enzyme_builder(cls, enzyme_builder: psc.PyEnzymeBuilder) -> 'EnzymeBuilder':
        instance = cls.__new__(cls)
        instance.__enzyme_builder_ptr = enzyme_builder
        return instance

    def get_enzyme_parameters(self) -> EnzymeParameters:
        return EnzymeParameters.from_py_enzyme_parameters(self.__enzyme_builder_ptr.get_enzyme_parameters())

    def get_py_ptr(self):
        return self.__enzyme_builder_ptr

    @property
    def missed_cleavages(self) -> int:
        return self.__enzyme_builder_ptr.missed_cleavages

    @property
    def min_len(self) -> int:
        return self.__enzyme_builder_ptr.min_len

    @property
    def max_len(self) -> int:
        return self.__enzyme_builder_ptr.max_len

    @property
    def cleave_at(self) -> str:
        return self.__enzyme_builder_ptr.cleave_at

    @property
    def restrict(self) -> str:
        return self.__enzyme_builder_ptr.restrict

    @property
    def c_terminal(self) -> bool:
        return self.__enzyme_builder_ptr.c_terminal

    @property
    def semi_enzymatic(self) -> bool:
        return self.__enzyme_builder_ptr.semi_enzymatic

    def __repr__(self):
        return f"EnzymeBuilder(missed_cleavages: {self.missed_cleavages}, min_len: {self.min_len}, " \
               f"max_len: {self.max_len}, cleave_at: {self.cleave_at}, restrict: {self.restrict}, " \
               f"c_terminal: {self.c_terminal}, semi_enzymatic: {self.semi_enzymatic})"


class SageSearchConfiguration:
    def __init__(self,
                 fasta: str,
                 bucket_size: int = 8192,
                 enzyme_builder: 'EnzymeBuilder' = EnzymeBuilder.default_trypsin(),
                 fragment_min_mz: float = 150,
                 fragment_max_mz: float = 2_000,
                 peptide_min_mass: float = 500,
                 peptide_max_mass: float = 5_000,
                 ion_kinds: List[IonType] = None,
                 min_ion_index: int = 2,
                 static_mods: Dict[ModificationSpecificity, float] = None,
                 variable_mods: Dict[ModificationSpecificity, List[float]] = None,
                 max_variable_mods: int = 2,
                 decoy_tag: str = 'rev_',
                 generate_decoys: bool = True,
                 shuffle_decoys: Union[bool, None] = None,
                 keep_ends: Union[bool, None] = None,
                 ):
        """SageSearchConfiguration class

        Args:
            fasta (str): The fasta file
            bucket_size (int, optional): The bucket size. Defaults to 8192.
            enzyme_builder (EnzymeBuilder, optional): The enzyme builder. Defaults to EnzymeBuilder.default_trypsin().
            fragment_min_mz (float, optional): The minimum fragment m/z. Defaults to 150.
            fragment_max_mz (float, optional): The maximum fragment m/z. Defaults to 2000.
            peptide_min_mass (float, optional): The minimum peptide mass. Defaults to 500.
            peptide_max_mass (float, optional): The maximum peptide mass. Defaults to 5000.
            ion_kinds (List[IonType], optional): The ion types. Defaults to None.
            min_ion_index (int, optional): The minimum ion index. Defaults to 2.
            static_mods (Dict[ModificationSpecificity, float], optional): The static modifications. Defaults to None.
            variable_mods (Dict[ModificationSpecificity, List[float]], optional): The variable modifications. Defaults to None.
            max_variable_mods (int, optional): The maximum number of variable modifications. Defaults to 2.
            decoy_tag (str, optional): The decoy tag. Defaults to 'rev_'.
            generate_decoys (bool, optional): Whether to generate decoys. Defaults to True.
            shuffle_decoys (Union[bool, None], optional): Whether to shuffle decoys. Defaults to None.
            keep_ends (Union[bool, None], optional): Whether to include start and end amino acid for permutation strategy. Defaults to None.
        """
        self.__py_parameter_ptr = psc.PyParameters(
            find_next_power_of_2(bucket_size),
            enzyme_builder.get_py_ptr(),
            fragment_min_mz,
            fragment_max_mz,
            peptide_min_mass,
            peptide_max_mass,
            min_ion_index,
            {k.get_py_ptr(): v for k, v in static_mods.items()} if static_mods is not None else {},
            {k.get_py_ptr(): v for k, v in variable_mods.items()} if variable_mods is not None else {},
            max_variable_mods,
            decoy_tag,
            generate_decoys,
            fasta,
            ion_kinds,
            shuffle_decoys,
            keep_ends
        )

    @classmethod
    def from_py_parameters(cls, parameters: psc.PyParameters) -> 'SageSearchConfiguration':
        instance = cls.__new__(cls)
        instance.__py_parameter_ptr = parameters
        return instance

    def get_py_ptr(self):
        return self.__py_parameter_ptr

    def _digest(self):
        return [Peptide.from_py_peptide(p) for p in self.__py_parameter_ptr.digest()]

    def generate_indexed_database(self) -> 'IndexedDatabase':
        """Generate the indexed database

        Returns:
            IndexedDatabase: The indexed database
        """
        return IndexedDatabase.from_py_indexed_database(self.__py_parameter_ptr.build_indexed_database())

    @property
    def bucket_size(self):
        return self.__py_parameter_ptr.bucket_size

    @property
    def enzyme_builder(self):
        return EnzymeBuilder.from_py_enzyme_builder(self.__py_parameter_ptr.enzyme_builder)

    @property
    def fragment_min_mz(self):
        return self.__py_parameter_ptr.fragment_min_mz

    @property
    def fragment_max_mz(self):
        return self.__py_parameter_ptr.fragment_max_mz

    @property
    def peptide_min_mass(self):
        return self.__py_parameter_ptr.peptide_min_mass

    @property
    def peptide_max_mass(self):
        return self.__py_parameter_ptr.peptide_max_mass

    @property
    def ion_kinds(self):
        return [IonType.from_py_kind(k) for k in self.__py_parameter_ptr.ion_kinds]

    @property
    def min_ion_index(self):
        return self.__py_parameter_ptr.min_ion_index

    @property
    def static_mods(self):
        return {ModificationSpecificity.from_py_modification_specificity(k): v for k, v in self.__py_parameter_ptr.static_mods.items()}

    @property
    def variable_mods(self):
        return {ModificationSpecificity.from_py_modification_specificity(k): v for k, v in self.__py_parameter_ptr.variable_mods.items()}

    @property
    def max_variable_mods(self):
        return self.__py_parameter_ptr.max_variable_mods

    @property
    def decoy_tag(self):
        return self.__py_parameter_ptr.decoy_tag

    @property
    def generate_decoys(self):
        return self.__py_parameter_ptr.generate_decoys

    @property
    def fasta(self):
        return self.__py_parameter_ptr.fasta

    def __repr__(self):
        return f"SageSearchConfiguration(bucket_size: {self.bucket_size}, enzyme_builder: {self.enzyme_builder}, " \
               f"fragment_min_mz: {self.fragment_min_mz}, fragment_max_mz: {self.fragment_max_mz}, " \
               f"peptide_min_mass: {self.peptide_min_mass}, peptide_max_mass: {self.peptide_max_mass}, " \
               f"ion_kinds: {self.ion_kinds}, min_ion_index: {self.min_ion_index}, " \
               f"max_variable_mods: {self.max_variable_mods}, decoy_tag: {self.decoy_tag}, " \
               f"generate_decoys: {self.generate_decoys})"


class IndexedDatabase:
    def __init__(self, peptides: List[Peptide], fragments: List[Theoretical], ion_kinds: List[IonType],
                 min_value: List[float],
                 potential_mods: List[Tuple[(ModificationSpecificity, float)]], bucket_size: int,
                 generate_decoys: bool, decoy_tag: str):
        """IndexedDatabase class

        Args:
            peptides (List[Peptide]): The peptides in the database
            fragments (List[Theoretical]): The theoretical fragments in the database
            ion_kinds (List[IonType]): The ion types in the database
            min_value (List[float]): The minimum values in the database
            potential_mods (List[Tuple[(ModificationSpecificity, float)]]): The potential modifications in the database
            bucket_size (int): The bucket size
            generate_decoys (bool): Whether to generate decoys
            decoy_tag (str): The decoy tag
        """
        self.__indexed_database_ptr = psc.PyIndexedDatabase(
            [p.get_py_ptr() for p in peptides],
            [f.get_py_ptr() for f in fragments],
            [k.get_py_ptr() for k in ion_kinds],
            min_value,
            [(k.get_py_ptr(), v) for (k, v) in potential_mods],
            bucket_size,
            generate_decoys,
            decoy_tag,
        )

    @classmethod
    def from_py_indexed_database(cls, indexed_database: psc.PyIndexedDatabase):
        instance = cls.__new__(cls)
        instance.__indexed_database_ptr = indexed_database
        return instance

    def get_py_ptr(self):
        return self.__indexed_database_ptr

    def __getitem__(self, item: Union[int, PeptideIx]) -> Peptide:
        if isinstance(item, int):
            pep_ix = PeptideIx(item)
            return Peptide.from_py_peptide(self.__indexed_database_ptr[pep_ix.get_py_ptr()])
        elif isinstance(item, PeptideIx):
            return Peptide.from_py_peptide(self.__indexed_database_ptr[item.get_py_ptr()])
        else:
            raise ValueError(f"Invalid item type: {type(item)}")

    def query(self, precursor_mass: float, precursor_tolerance: Tolerance, fragment_tolerance: Tolerance):
        return IndexedQuery.from_py_indexed_query(self.__indexed_database_ptr.query(precursor_mass,
                                                                                    precursor_tolerance.get_py_ptr(),
                                                                                    fragment_tolerance.get_py_ptr()))

    @property
    def _peptides(self):
        """Peptides in the database
        CAUTION: This method is for debugging purposes only. Do not use in production, RAM usage will be very high.
        Returns:
            List[Peptide]: The peptides in the database
        """
        return [Peptide.from_py_peptide(p) for p in self.__indexed_database_ptr.peptides]

    def peptides_as_string(self):
        return np.array(self.__indexed_database_ptr.peptides_as_string())

    def mono_masses(self):
        return np.array(self.__indexed_database_ptr.mono_masses())

    def modifications(self):
        return self.__indexed_database_ptr.modifications()

    @property
    def df(self) -> pd.DataFrame:
        """Peptides in the database as a pandas DataFrame

        Returns:
            pd.DataFrame: The peptides in the database as a pandas DataFrame
        """

        data = {
            'sequence': self.peptides_as_string(),
            'monoisotopic_mass': self.mono_masses()
        }

        return pd.DataFrame(data)

    @property
    def num_peptides(self) -> int:
        """Number of peptides in the database

        Returns:
            int: Number of peptides in the database
        """
        return self.__indexed_database_ptr.num_peptides

    @property
    def _fragments(self):
        """Fragments in the database
        CAUTION: This method is for debugging purposes only. Do not use in production, RAM usage will be very high.
        Returns:
            List[Theoretical]: The fragments in the database
        """
        return [Theoretical.from_py_theoretical(f) for f in self.__indexed_database_ptr.fragments]

    @property
    def fragment_indices(self):
        return self.__indexed_database_ptr.fragment_indices

    @property
    def fragment_mzs(self):
        return self.__indexed_database_ptr.fragment_mzs

    def fragment_dict(self):
        return self.__indexed_database_ptr.fragment_dict()

    @property
    def num_fragments(self) -> int:
        """Number of fragments in the database

        Returns:
            int: Number of fragments in the database
        """
        return self.__indexed_database_ptr.num_fragments

    @property
    def ion_kinds(self):
        return [IonType.from_py_kind(k) for k in self.__indexed_database_ptr.ion_kinds]

    @property
    def min_value(self):
        return self.__indexed_database_ptr.min_value

    @property
    def potential_mods(self):
        return [(ModificationSpecificity.from_py_modification_specificity(k), v) for (k, v) in self.__indexed_database_ptr.potential_mods]

    @property
    def bucket_size(self):
        return self.__indexed_database_ptr.bucket_size

    @property
    def generate_decoys(self):
        return self.__indexed_database_ptr.generate_decoys

    @property
    def decoy_tag(self):
        return self.__indexed_database_ptr.decoy_tag

    def __repr__(self):
        return f"IndexedDatabase(peptides: {self.num_peptides}, fragments: {self.num_fragments}, ion_kinds: {self.ion_kinds}, " \
               f"num_buckets: {len(self.min_value)}, " \
               f"bucket_size: {self.bucket_size}, generate_decoys: {self.generate_decoys}, " \
               f"decoy_tag: {self.decoy_tag})"


class IndexedQuery:
    def __init__(self, precursor_mass: float, precursor_tolerance: Tolerance, fragment_tolerance: Tolerance,
                 pre_idx_lo: int, pre_idx_hi: int):
        """IndexedQuery class

        Args:
            precursor_mass (float): The precursor mass
            precursor_tolerance (Tolerance): The precursor tolerance
            fragment_tolerance (Tolerance): The fragment tolerance
            pre_idx_lo (int): The lower index
            pre_idx_hi (int): The upper index
        """
        self.__indexed_query_ptr = psc.PyIndexedQuery(precursor_mass, precursor_tolerance.get_py_ptr(),
                                                      fragment_tolerance.get_py_ptr(), pre_idx_lo, pre_idx_hi)

    @classmethod
    def from_py_indexed_query(cls, indexed_query: psc.PyIndexedQuery):
        instance = cls.__new__(cls)
        instance.__indexed_query_ptr = indexed_query
        return instance

    def get_py_ptr(self):
        return self.__indexed_query_ptr

    @property
    def precursor_mass(self):
        return self.__indexed_query_ptr.precursor_mass

    @property
    def precursor_tolerance(self):
        return Tolerance.from_py_tolerance(self.__indexed_query_ptr.precursor_tolerance)

    @property
    def fragment_tolerance(self):
        return Tolerance.from_py_tolerance(self.__indexed_query_ptr.fragment_tolerance)

    @property
    def pre_idx_lo(self):
        return self.__indexed_query_ptr.pre_idx_lo

    @property
    def pre_idx_hi(self):
        return self.__indexed_query_ptr.pre_idx_hi

    def __repr__(self):
        return f"IndexedQuery(precursor_mass: {self.precursor_mass}, precursor_tolerance: {self.precursor_tolerance}, " \
               f"fragment_tolerance: {self.fragment_tolerance}, pre_idx_lo: {self.pre_idx_lo}, " \
               f"pre_idx_hi: {self.pre_idx_hi})"


def find_next_power_of_2(n):
    """
    Find the next power of two greater than or equal to n.

    Args:
    n (int): A positive integer.

    Returns:
    int: The smallest power of two greater than or equal to n.
    """
    return 1 << (n-1).bit_length()



from .scoring import Scorer, Fragments, IonType, PeptideSpectrumMatch, Feature
from .database import IndexedDatabase, EnzymeBuilder, SageSearchConfiguration
from .spectrum import RawSpectrum, ProcessedSpectrum, Precursor, SpectrumProcessor, Representation
from .mass import Tolerance
from .ion_series import IonSeries
from .peptide import Peptide
from .modification import SAGE_KNOWN_MODS, validate_mods, validate_var_mods



from typing import List
from sagepy.core.database import PeptideIx
import sagepy_connector
psc = sagepy_connector.py_lfq


class PeakScoringStrategy:
    """PeakScoringStrategy class

    Args:
        strategy (str): The peak scoring strategy, allowed values are: retention_time, spectral_angle, intensity, hybrid
    """
    def __init__(self, strategy: str):
        strategies = ["retention_time", "spectral_angle", "intensity", "hybrid"]
        if strategy in strategies:
            self.__peak_scoring_strategy_ptr = psc.PyPeakScoringStrategy(strategy)
        else:
            raise ValueError(f"Invalid peak scoring strategy, allowed values are: {strategies}")

    @classmethod
    def from_py_peak_scoring_strategy(cls, peak_scoring_strategy: psc.PyPeakScoringStrategy):
        instance = cls.__new__(cls)
        instance.__peak_scoring_strategy_ptr = peak_scoring_strategy
        return instance

    @property
    def strategy(self):
        return self.__peak_scoring_strategy_ptr.strategy

    def __repr__(self):
        return f"PeakScoringStrategy({self.__peak_scoring_strategy_ptr.strategy})"

    def get_py_ptr(self):
        return self.__peak_scoring_strategy_ptr


class IntegrationStrategy:
    def __init__(self, strategy: str):
        strategies = ['apex', 'sum']
        if strategy in strategies:
            self.__integration_strategy_ptr = psc.PyIntegrationStrategy(strategy)
        else:
            raise ValueError(f"Invalid integration strategy, allowed values are: {strategies}")

    @classmethod
    def from_py_integration_strategy(cls, integration_strategy: psc.PyIntegrationStrategy):
        instance = cls.__new__(cls)
        instance.__integration_strategy_ptr = integration_strategy
        return instance

    @property
    def strategy(self):
        return self.__integration_strategy_ptr.strategy

    def __repr__(self):
        return f"IntegrationStrategy({self.__integration_strategy_ptr.strategy})"

    def get_py_ptr(self):
        return self.__integration_strategy_ptr


class PrecursorId:
    def __init__(self, peptide_id: PeptideIx):
        self.__precursor_id_ptr = psc.PyPrecursorId(peptide_id.get_py_ptr())

    @classmethod
    def from_py_precursor_id(cls, precursor_id: psc.PyPrecursorId):
        instance = cls.__new__(cls)
        instance.__precursor_id_ptr = precursor_id
        return instance

    @classmethod
    def from_charged(cls, peptide_id: PeptideIx, charge: int) -> 'PrecursorId':
        return cls.from_py_precursor_id(psc.PyPrecursorId.from_charged(peptide_id.get_py_ptr(), charge))

    def get_ptr(self):
        return self.__precursor_id_ptr


class LfqSettings:
    # TODO: check if tolerance should be of type Tolerance instead of float
    def __init__(self, peak_scoring_strategy: PeakScoringStrategy, integration_strategy: IntegrationStrategy,
                 spectral_angle: float, ppm_tolerance: float, combine_charge_states: bool):
        self.__lfq_settings_ptr = psc.PyLfqSettings(peak_scoring_strategy.get_py_ptr(),
                                                    integration_strategy.get_py_ptr(),
                                                    spectral_angle, ppm_tolerance, combine_charge_states)

    @classmethod
    def from_py_lfq_settings(cls, lfq_settings: psc.PyLfqSettings):
        instance = cls.__new__(cls)
        instance.__lfq_settings_ptr = lfq_settings
        return instance

    def get_py_ptr(self):
        return self.__lfq_settings_ptr

    @property
    def peak_scoring_strategy(self) -> PeakScoringStrategy:
        return PeakScoringStrategy.from_py_peak_scoring_strategy(self.__lfq_settings_ptr.peak_scoring_strategy)

    @property
    def integration_strategy(self) -> IntegrationStrategy:
        return IntegrationStrategy.from_py_integration_strategy(self.__lfq_settings_ptr.integration_strategy)

    @property
    def spectral_angle(self) -> float:
        return self.__lfq_settings_ptr.spectral_angle

    @property
    def ppm_tolerance(self) -> float:
        return self.__lfq_settings_ptr.ppm_tolerance

    @property
    def combine_charge_states(self) -> bool:
        return self.__lfq_settings_ptr.combine_charge_states

    def __repr__(self):
        return (f"LfqSettings(peak_scoring_strategy: {self.peak_scoring_strategy}, "
                f"integration_strategy: {self.integration_strategy}, "
                f"spectral_angle: {self.spectral_angle}, ppm_tolerance: {self.ppm_tolerance}, "
                f"combine_charge_states: {self.combine_charge_states})")


class PrecursorRange:
    def __init__(self, rt: float, mass_lo: float, mass_hi: float, charge: int,
                 isotope: int, peptide: PeptideIx, file_id: int, decoy: bool):
        self.__precursor_range_ptr = psc.PyPrecursorRange(rt, mass_lo, mass_hi, charge, isotope,
                                                          peptide.get_py_ptr(), file_id, decoy)

    @classmethod
    def from_py_precursor_range(cls, precursor_range: psc.PyPrecursorRange):
        instance = cls.__new__(cls)
        instance.__precursor_range_ptr = precursor_range
        return instance

    def get_py_ptr(self):
        return self.__precursor_range_ptr

    @property
    def rt(self) -> float:
        return self.__precursor_range_ptr.rt

    @property
    def mass_lo(self) -> float:
        return self.__precursor_range_ptr.mass_lo

    @property
    def mass_hi(self) -> float:
        return self.__precursor_range_ptr.mass_hi

    @property
    def charge(self) -> int:
        return self.__precursor_range_ptr.charge

    @property
    def isotope(self) -> int:
        return self.__precursor_range_ptr.isotope

    @property
    def peptide(self) -> PeptideIx:
        return PeptideIx.from_py_peptide_ix(self.__precursor_range_ptr.peptide)

    @property
    def file_id(self) -> int:
        return self.__precursor_range_ptr.file_id

    @property
    def decoy(self) -> bool:
        return self.__precursor_range_ptr.decoy

    def __repr__(self):
        return (f"PrecursorRange(rt: {self.rt}, mass_lo: {self.mass_lo}, mass_hi: {self.mass_hi}, "
                f"charge: {self.charge}, isotope: {self.isotope}, peptide: {self.peptide}, "
                f"file_id: {self.file_id}, decoy: {self.decoy})")


class FeatureMap:
    def __init__(self, ranges: List[PrecursorRange], min_rts: List[float], bin_size: int, settings: LfqSettings):
        self.__feature_map_ptr = psc.PyFeatureMap([
            r.get_py_ptr() for r in ranges
        ], min_rts, bin_size, settings.get_py_ptr())

    @classmethod
    def from_py_feature_map(cls, feature_map: psc.PyFeatureMap):
        instance = cls.__new__(cls)
        instance.__feature_map_ptr = feature_map
        return instance

    def get_py_ptr(self):
        return self.__feature_map_ptr

    @property
    def _ranges(self) -> List[PrecursorRange]:
        return [PrecursorRange.from_py_precursor_range(r) for r in self.__feature_map_ptr.ranges]

    @property
    def min_rts(self) -> List[float]:
        return self.__feature_map_ptr.min_rts

    @property
    def bin_size(self) -> int:
        return self.__feature_map_ptr.bin_size

    @property
    def settings(self) -> LfqSettings:
        return LfqSettings.from_py_lfq_settings(self.__feature_map_ptr.settings)

    def get_num_ranges(self) -> int:
        return self.__feature_map_ptr.get_num_ranges()

    def __repr__(self):
        return f"FeatureMap(num_ranges: {self.get_num_ranges()}, bin_size: {self.bin_size}, settings: {self.settings})"


class Query:
    def __init__(self, ranges: List[PrecursorRange], page_lo: int,
                 page_hi: int, bin_size: int, min_rt: float, max_rt: float):
        self.__query_ptr = psc.PyQuery([
            r.get_py_ptr() for r in ranges
        ], page_lo, page_hi, bin_size, min_rt, max_rt)

    @classmethod
    def from_py_query(cls, query: psc.PyQuery):
        instance = cls.__new__(cls)
        instance.__query_ptr = query
        return instance

    def get_py_ptr(self):
        return self.__query_ptr

    @property
    def _ranges(self) -> List[PrecursorRange]:
        return [PrecursorRange.from_py_precursor_range(r) for r in self.__query_ptr.ranges]

    @property
    def page_lo(self) -> int:
        return self.__query_ptr.page_lo

    @property
    def page_hi(self) -> int:
        return self.__query_ptr.page_hi

    @property
    def bin_size(self) -> int:
        return self.__query_ptr.bin_size

    @property
    def min_rt(self) -> float:
        return self.__query_ptr.min_rt

    @property
    def max_rt(self) -> float:
        return self.__query_ptr.max_rt

    def get_num_ranges(self) -> int:
        return self.__query_ptr.get_num_ranges()

    def __repr__(self):
        return f"Query(num_ranges: {self.get_num_ranges()}, page_lo: {self.page_lo}, page_hi: {self.page_hi}, " \
                f"bin_size: {self.bin_size}, min_rt: {self.min_rt}, max_rt: {self.max_rt})"
